from __future__ import annotations

from datetime import datetime
import logging
from types import MappingProxyType
from typing import (Any,
                    Callable,
                    Dict,
                    Mapping,
                    Optional,
                    Protocol,
                    runtime_checkable,
                    Tuple,
                    Type,
                    TypeVar)

from omnipy.api.enums import (ConfigPersistOutputsOptions,
                              ConfigRestoreOutputsOptions,
                              EngineChoice,
                              PersistOutputsOptions,
                              RestoreOutputsOptions,
                              RunState)


class IsNestedContext(Protocol):
    def __enter__(self):
        ...

    def __exit__(self, exc_type, exc_value, traceback):
        ...


class IsJobConfigHolder(Protocol):
    engine: Optional[IsEngine]
    config: Optional[IsJobConfig]

    def set_config(self, config: IsJobConfig) -> None:
        ...

    def set_engine(self, engine: IsEngine) -> None:
        ...


class IsJobCreator(IsNestedContext, IsJobConfigHolder, Protocol):
    nested_context_level: int
    time_of_cur_toplevel_nested_context_run: datetime


class IsJob(Protocol):
    name: str
    unique_name: str
    config: Optional[IsJobConfig]
    in_flow_context: bool
    time_of_cur_toplevel_flow_run: datetime

    @classmethod
    def create(cls, *init_args: object, **init_kwargs: object) -> 'IsJob':
        ...

    def revise(self) -> 'IsJobTemplate':
        ...

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        ...

    def regenerate_unique_name(self) -> None:
        ...

    def __eq__(self, other: object):
        ...


class IsJobTemplate(Protocol):
    config: Optional[IsJobConfig]
    engine: Optional[IsTaskRunnerEngine]
    in_flow_context: bool

    @classmethod
    def create(cls, *init_args: object, **init_kwargs: object) -> 'IsJobTemplate':
        ...

    def run(self, *args: object, **kwargs: object) -> object:
        ...

    def apply(self) -> IsJob:
        ...

    def refine(self, *args: object, update: bool = True, **kwargs: object) -> 'IsJobTemplate':
        ...

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        ...


class IsFuncJob(IsJob, Protocol):
    param_signatures: MappingProxyType
    return_type: Type[object]
    iterate_over_data_files: bool
    fixed_params: MappingProxyType[str, object]
    param_key_map: MappingProxyType[str, str]
    result_key: Optional[str]
    persist_outputs: Optional[PersistOutputsOptions]
    restore_outputs: Optional[RestoreOutputsOptions]
    will_persist_outputs: PersistOutputsOptions
    will_restore_outputs: RestoreOutputsOptions

    def has_coroutine_func(self) -> bool:
        ...

    def get_call_args(self, *args: object, **kwargs: object) -> Dict[str, object]:
        ...


class IsFuncJobTemplate(IsJobTemplate, Protocol):
    name: str
    unique_name: str
    param_signatures: MappingProxyType
    return_type: Type[object]
    iterate_over_data_files: bool
    fixed_params: MappingProxyType[str, object]
    param_key_map: MappingProxyType[str, str]
    result_key: Optional[str]
    persist_outputs: Optional[PersistOutputsOptions]
    restore_outputs: Optional[RestoreOutputsOptions]
    will_persist_outputs: PersistOutputsOptions
    will_restore_outputs: RestoreOutputsOptions

    def apply(self) -> IsFuncJob:
        ...

    def refine(self,
               update: bool = True,
               name: Optional[str] = None,
               iterate_over_data_files: bool = False,
               fixed_params: Optional[Mapping[str, object]] = None,
               param_key_map: Optional[Mapping[str, str]] = None,
               result_key: Optional[str] = None,
               persist_outputs: Optional[PersistOutputsOptions] = None,
               restore_outputs: Optional[RestoreOutputsOptions] = None,
               **kwargs: object) -> 'IsFuncJobTemplate':
        ...

    def regenerate_unique_name(self) -> None:
        ...


class IsTask(IsFuncJob, Protocol):
    ...


class IsTaskTemplate(IsFuncJobTemplate, Protocol):
    def apply(self) -> IsTask:
        ...


class IsFlow(IsFuncJob, Protocol):
    flow_context: IsNestedContext
    time_of_last_run: datetime


class IsFlowTemplate(IsFuncJobTemplate, Protocol):
    def apply(self) -> IsFlow:
        ...


class IsTaskTemplatesFlow(IsFlow, Protocol):
    task_templates: Tuple[IsTaskTemplate, ...]

    def _accept_call_func_decorator(self, call_func_decorator: GeneralDecorator) -> None:
        ...


class IsTaskTemplatesFlowTemplate(IsFuncJobTemplate, Protocol):
    task_templates: Tuple[IsTaskTemplate, ...]

    def refine(self,
               *task_templates: IsTaskTemplate,
               update: bool = True,
               name: Optional[str] = None,
               iterate_over_data_files: bool = False,
               fixed_params: Optional[Mapping[str, object]] = None,
               param_key_map: Optional[Mapping[str, str]] = None,
               result_key: Optional[str] = None,
               persist_outputs: Optional[PersistOutputsOptions] = None,
               restore_outputs: Optional[RestoreOutputsOptions] = None,
               **kwargs: object) -> 'IsTaskTemplatesFlowTemplate':
        ...


class IsLinearFlow(IsTaskTemplatesFlow, Protocol):
    ...


class IsLinearFlowTemplate(IsLinearFlow, IsTaskTemplatesFlowTemplate, Protocol):
    def apply(self) -> IsLinearFlow:
        ...


class IsDagFlow(IsTaskTemplatesFlow, Protocol):
    ...


class IsDagFlowTemplate(IsDagFlow, IsTaskTemplatesFlowTemplate, Protocol):
    def apply(self) -> IsDagFlow:
        ...


class IsFuncFlow(IsFlow, Protocol):
    ...


class IsFuncFlowTemplate(IsFuncFlow, IsFlowTemplate, Protocol):
    def apply(self) -> IsFuncFlow:
        ...


JobBaseT = TypeVar('JobBaseT', bound='JobBase', covariant=True)
JobT = TypeVar('JobT', bound='Job', covariant=True)
JobTemplateT = TypeVar('JobTemplateT', bound='JobTemplate', covariant=True)
FuncJobTemplateT = TypeVar('FuncJobTemplateT', bound='FuncJobTemplate', covariant=True)


class IsFuncJobTemplateCallable(Protocol[FuncJobTemplateT]):
    def __call__(
        self,
        name: Optional[str] = None,
        iterate_over_data_files: bool = False,
        fixed_params: Optional[Mapping[str, object]] = None,
        param_key_map: Optional[Mapping[str, str]] = None,
        result_key: Optional[str] = None,
        persist_outputs: Optional[PersistOutputsOptions] = None,
        restore_outputs: Optional[RestoreOutputsOptions] = None,
        **kwargs: object,
    ) -> Callable[[Callable], FuncJobTemplateT]:
        ...


TaskTemplatesFlowTemplateT = TypeVar(
    'TaskTemplatesFlowTemplateT', bound='TaskTemplatesFlowTemplate', covariant=True)


class IsTaskTemplatesFlowTemplateCallable(Protocol[TaskTemplatesFlowTemplateT]):
    def __call__(
        self,
        *task_templates: 'TaskTemplate',
        name: Optional[str] = None,
        iterate_over_data_files: bool = False,
        fixed_params: Optional[Mapping[str, object]] = None,
        param_key_map: Optional[Mapping[str, str]] = None,
        result_key: Optional[str] = None,
        persist_outputs: Optional[PersistOutputsOptions] = None,
        restore_outputs: Optional[RestoreOutputsOptions] = None,
        **kwargs: object,
    ) -> Callable[[Callable], TaskTemplatesFlowTemplateT]:
        ...


FlowT = TypeVar('FlowT', bound='Flow', covariant=True)
FlowBaseT = TypeVar('FlowBaseT', bound='FlowBase', covariant=True)
FlowTemplateT = TypeVar('FlowTemplateT', bound='FlowTemplate', covariant=True)
GeneralDecorator = Callable[[Callable], Callable]


@runtime_checkable
class IsEngine(Protocol):
    def __init__(self) -> None:
        ...

    @classmethod
    def get_config_cls(cls) -> Type[IsEngineConfig]:
        ...

    def set_config(self, config: IsEngineConfig) -> None:
        ...

    def set_registry(self, registry: Optional['IsRunStateRegistry']) -> None:
        ...


@runtime_checkable
class IsTaskRunnerEngine(IsEngine, Protocol):
    def apply_task_decorator(self, task: IsTask, job_callback_accept_decorator: Callable) -> None:
        ...


@runtime_checkable
class IsLinearFlowRunnerEngine(IsEngine, Protocol):
    def apply_linear_flow_decorator(self,
                                    linear_flow: IsLinearFlow,
                                    job_callback_accept_decorator: Callable) -> None:
        ...


@runtime_checkable
class IsDagFlowRunnerEngine(IsEngine, Protocol):
    def apply_dag_flow_decorator(self, dag_flow: IsDagFlow,
                                 job_callback_accept_decorator: Callable) -> None:
        ...


@runtime_checkable
class IsFuncFlowRunnerEngine(IsEngine, Protocol):
    def apply_func_flow_decorator(self,
                                  func_flow: IsFuncFlow,
                                  job_callback_accept_decorator: Callable) -> None:
        ...


class IsRunStateRegistry(Protocol):
    def __init__(self) -> None:
        ...

    def get_job_state(self, job: IsJob) -> RunState:
        ...

    def get_job_state_datetime(self, job: IsJob, state: RunState) -> datetime:
        ...

    def all_jobs(self, state: Optional[RunState] = None) -> Tuple[IsJob, ...]:  # noqa
        ...

    def set_job_state(self, job: IsJob, state: RunState) -> None:
        ...

    def set_logger(self, logger: Optional[logging.Logger]) -> None:
        ...

    def set_config(self, config: IsRunStateRegistryConfig) -> None:
        ...


class IsEngineConfig(Protocol):
    ...


class IsLocalRunnerConfig(IsEngineConfig, Protocol):
    ...


class IsPrefectEngineConfig(IsEngineConfig, Protocol):
    use_cached_results: int = False


class IsRunStateRegistryConfig(Protocol):
    verbose: bool = True

    def __init__(
            self,
            verbose: bool = True,  # noqa
            *args: Any,
            **kwargs: Any) -> None:
        ...


class IsConfigPublisher(Protocol):
    def subscribe(self, config_item: str, callback_fun: Callable[[Any], None]):
        ...

    def unsubscribe_all(self) -> None:
        ...


class IsJobConfig(Protocol):
    persist_outputs: ConfigPersistOutputsOptions
    restore_outputs: ConfigRestoreOutputsOptions
    persist_data_dir_path: str


class IsRuntimeConfig(IsConfigPublisher, Protocol):
    job: IsJobConfig
    engine: EngineChoice
    local: IsLocalRunnerConfig
    prefect: IsPrefectEngineConfig
    registry: IsRunStateRegistryConfig

    def __init__(
            self,
            job: Optional[IsJobConfig] = None,  # noqa
            engine: EngineChoice = EngineChoice.LOCAL,  # noqa
            local_config: Optional[IsLocalRunnerConfig] = None,  # noqa
            prefect_config: Optional[IsPrefectEngineConfig] = None,  # noqa
            registry_config: Optional[IsRunStateRegistryConfig] = None,  # noqa
            *args: Any,
            **kwargs: Any) -> None:
        ...


class IsRuntimeObjects(IsConfigPublisher, Protocol):
    logger: logging.Logger
    registry: IsRunStateRegistry
    job_creator: IsJobConfigHolder
    local: IsEngine
    prefect: IsEngine

    def __init__(
            self,
            logger: Optional[logging.Logger] = None,  # noqa
            registry: Optional[IsRunStateRegistry] = None,  # noqa
            job_creator: Optional[IsJobConfigHolder] = None,  # noqa
            local: Optional[IsEngine] = None,  # noqa
            prefect: Optional[IsEngine] = None,  # noqa
            *args: Any,
            **kwargs: Any) -> None:
        ...


class IsRuntime(IsConfigPublisher, Protocol):
    objects: IsRuntimeObjects
    config: IsRuntimeConfig

    def __init__(
            self,
            objects: Optional[IsRuntimeObjects] = None,  # noqa
            config: Optional[IsRuntimeConfig] = None,  # noqa
            *args: Any,
            **kwargs: Any) -> None:
        ...

    def reset_subscriptions(self):
        ...



from dataclasses import dataclass


@dataclass
class EngineConfig:
    ...


@dataclass
class LocalRunnerConfig(EngineConfig):
    ...


@dataclass
class PrefectEngineConfig(EngineConfig):
    use_cached_results: int = False



from io import StringIO
from typing import List, Optional

from omnipy.compute.task import TaskTemplate
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

from . import pd
from .models import ListOfPandasDatasetsWithSameNumberOfFiles, PandasDataset


@TaskTemplate()
def from_csv(dataset: Dataset[Model[bytes]],
             delimiter: str = ',',
             first_row_as_col_names=True,
             col_names: Optional[List[str]] = None,
             ignore_comments: bool = True,
             comments_char: str = '#') -> PandasDataset:
    out_dataset = PandasDataset()
    for key, item in dataset.items():
        df = pd.read_csv(
            StringIO(item),
            sep=delimiter,
            header='infer' if first_row_as_col_names else 0,
            names=col_names,
            comment=comments_char if ignore_comments else None,
            encoding='utf8',
        )
        out_dataset[key] = df
    return out_dataset


@TaskTemplate()
def to_csv(
    dataset: PandasDataset,
    delimiter: str = ',',
    first_row_as_col_names=True,
    col_names: Optional[List[str]] = None,
) -> Dataset[Model[str]]:
    out_dataset = Dataset[Model[str]]()
    for key, df in dataset.items():
        csv_stream = StringIO()
        df.to_csv(
            csv_stream,
            sep=delimiter,
            header=col_names if col_names else True if first_row_as_col_names else False,
            encoding='utf8',
            index=False)
        out_dataset[key] = csv_stream.getvalue()
    return out_dataset


@TaskTemplate()
def extract_columns_as_files(dataset: PandasDataset, col_names: List[str]) -> PandasDataset:
    out_dataset = PandasDataset()
    for key, item in dataset.items():
        df = dataset[key]
        out_dataset[key] = df.loc[:, ~df.columns.isin(col_names)]

        for col_name in col_names:
            out_dataset[f'{key}.{col_name}'] = pd.DataFrame(df[col_name])
    return out_dataset


@TaskTemplate()
def concat_dataframes_across_datasets(dataset_list: ListOfPandasDatasetsWithSameNumberOfFiles,
                                      vertical=True) -> PandasDataset:
    # We know from the data type that there are at least two datasets and that there is an equal
    # amount of files/DataFrames in each dataset. This simplifies implementation.

    out_dataset = PandasDataset()
    out_datafile_names = tuple(dataset_list[0].keys())
    for df_index in range(len(out_datafile_names)):
        print([tuple(dataset.keys())[df_index] for dataset in dataset_list])
        df = pd.concat([tuple(dataset.values())[df_index] for dataset in dataset_list],
                       axis=0 if vertical else 1)
        out_dataset[out_datafile_names[df_index]] = df
    return out_dataset



from __future__ import annotations

from abc import ABCMeta
from datetime import datetime
from typing import Optional

from omnipy.api.protocols import IsJobConfig, IsJobConfigHolder, IsTaskRunnerEngine


class JobCreator:
    def __init__(self) -> None:
        self._engine: Optional[IsTaskRunnerEngine] = None
        self._config: Optional[IsJobConfig] = None
        self._nested_context_level: int = 0
        self._time_of_cur_toplevel_nested_context_run: Optional[datetime] = None

    def set_engine(self, engine: IsTaskRunnerEngine) -> None:
        self._engine = engine

    def set_config(self, config: IsJobConfig) -> None:
        self._config = config

    def __enter__(self):
        if self._nested_context_level == 0:
            self._time_of_cur_toplevel_nested_context_run = datetime.now()

        self._nested_context_level += 1

    def __exit__(self, exc_type, exc_value, traceback):
        self._nested_context_level -= 1

        if self._nested_context_level == 0:
            self._time_of_cur_toplevel_nested_context_run = None

    @property
    def engine(self) -> Optional[IsTaskRunnerEngine]:
        return self._engine

    @property
    def config(self) -> Optional[IsJobConfig]:
        return self._config

    @property
    def nested_context_level(self) -> int:
        return self._nested_context_level

    @property
    def time_of_cur_toplevel_nested_context_run(self) -> datetime:
        return self._time_of_cur_toplevel_nested_context_run


class JobBaseMeta(ABCMeta):
    _job_creator: IsJobConfigHolder = JobCreator()

    @property
    def job_creator(self) -> IsJobConfigHolder:
        return self._job_creator

    @property
    def config(self) -> Optional[IsJobConfig]:
        return self.job_creator.config

    @property
    def engine(self) -> Optional[IsTaskRunnerEngine]:
        return self.job_creator.engine

    @property
    def nested_context_level(self) -> int:
        return self.job_creator.nested_context_level



from functools import update_wrapper
from types import MethodWrapperType
from typing import Callable, cast, Protocol, runtime_checkable, Type, TypeVar

# Types

DecoratorClassT = TypeVar('DecoratorClassT', covariant=True)


@runtime_checkable
class CallableParamAfterSelf(Protocol):
    def __call__(self, callable_arg: Callable, /, *args: object, **kwargs: object) -> None:
        ...


@runtime_checkable
class CallableClass(Protocol[DecoratorClassT]):
    def __call__(self, *args: object, **kwargs: object) -> Callable[[Callable], DecoratorClassT]:
        ...


def callable_decorator_cls(cls: Type[DecoratorClassT]) -> CallableClass[DecoratorClassT]:
    """
    "Meta-decorator" that allows any class to function as a decorator for a callable.

    The only requirements are that 1) the first argument after self of the __init__() method needs
    to be annotated as a callable, and 2) the class must not already be callable (have a __call__()
    method).

    Arguments and keyword arguments to the class decorator are supported.
    """
    if not isinstance(cls.__call__, MethodWrapperType):
        cls._wrapped_call: Callable = cast(Callable, cls.__call__)

    def _forward_call_to_obj_if_callable(self, *args: object,
                                         **kwargs: object) -> Type[DecoratorClassT]:
        """
        __call__ method at the class level which forward the call to instance-level call methods,
        if present (hardcoded as '_obj_call()'). This is needed due to the peculiarity that Python
        only looks up special methods (with double underscores) at the class level, and not at the
        instance level. Used in the decoration process to forward __call__ calls to the object level
        _obj_call() methods, if present.

        See: https://stackoverflow.com/q/33824228
        """
        if hasattr(self, '_obj_call'):
            return self._obj_call(*args, **kwargs)
        if hasattr(self, '_wrapped_call'):
            return self._wrapped_call(*args, **kwargs)
        raise TypeError("'{}' object is not callable".format(self.__class__.__name__))

    setattr(cls, '__call__', _forward_call_to_obj_if_callable)

    def _real_callable(arg: object) -> bool:
        """
        Helper method needed to ignore the class level __call__ method when checking whether the
        decorated class is callable.
        """
        if callable(arg):
            if hasattr(arg, '__call__'):
                if hasattr(arg.__call__, '__func__'):
                    # due to mypy bug: https://github.com/python/mypy/issues/14123
                    call_func_name = getattr(arg.__call__, '__func__').__name__

                    if call_func_name == _forward_call_to_obj_if_callable.__name__:
                        return False
            return True
        return False

    _wrapped_new: Callable = cls.__new__

    def _new_wrapper(cls, *args: object, **kwargs: object) -> DecoratorClassT:
        if _wrapped_new is object.__new__:
            obj = _wrapped_new(cls)
        else:
            obj = _wrapped_new(cls, *args, **kwargs)
            # setattr(cls, '__new__', _wrapped_new)

        _wrapped_init: CallableParamAfterSelf = cast(CallableParamAfterSelf, obj.__class__.__init__)

        # Wrapper method that replaces the __init__ method of the decorated class
        def _init_wrapper(self, *args: object, **kwargs: object) -> None:
            args_list = list(args)

            def _init(callable_arg: Callable) -> None:
                _wrapped_init(self, callable_arg, *args_list, **kwargs)
                update_wrapper(self, callable_arg, updated=[])

            if len(args_list) == 1 and _real_callable(args_list[0]):
                # Decorate the callable directly
                _callable_arg: Callable = cast(Callable, args_list[0])
                args_list.pop(0)
                _init(_callable_arg)
            else:
                # Add an instance-level _obj_call method, which are again callable by the
                # class-level __call__ method. When this method is called, the provided _callable_arg
                # is decorated.
                def _init_as_obj_call_method(
                        self, _callable_arg: Callable) -> Type[DecoratorClassT]:  # noqa
                    _init(_callable_arg)
                    del self._obj_call
                    return self

                self._obj_call = _init_as_obj_call_method.__get__(self)

            setattr(self.__class__, '__init__', _wrapped_init)

        setattr(obj.__class__, '__init__', _init_wrapper)

        return obj

    setattr(cls, '__new__', _new_wrapper)

    return cast(CallableClass[DecoratorClassT], cls)



from typing import Iterable

from pydantic import constr, PositiveInt

from omnipy.compute.task import TaskTemplate

from ..json.models import JsonDataset
from .functions import encode_api


@TaskTemplate()
def import_dataset_from_encode(endpoints: Iterable[constr(min_length=1)],
                               max_data_item_count: PositiveInt) -> JsonDataset:
    dataset = JsonDataset()
    for endpoint in endpoints:
        dataset[endpoint] = encode_api(
            endpoint,
            limit=str(max_data_item_count),
            format='json',
        )
    return dataset



from __future__ import annotations

from collections import UserDict
import json
from typing import Any, Dict, Generic, Iterator, Tuple, Type, TypeVar, Union

# from orjson import orjson
from pydantic import Field, PrivateAttr, ValidationError
from pydantic.generics import GenericModel
from pydantic.utils import lenient_issubclass

from omnipy.data.model import Model

ModelT = TypeVar('ModelT', bound=Model)
Undefined = object()
DATA_KEY = 'data'

# def orjson_dumps(v, *, default):
#     # orjson.dumps returns bytes, to match standard json.dumps we need to decode
#     return orjson.dumps(v, default=default).decode()


class Dataset(GenericModel, Generic[ModelT], UserDict):
    """
    Dataset is a generic class. Subclasses of Dataset need to specify the type of the contents
    that is accepted according to its model.

    This can be done directly, e.g.:

    class MyDataset(Dataset[Dict[str, List[int]]):
        pass

    or indirectly through the use of Model, e.g.:

    class MyModel(Model[Dict[str, List[int]]):
        pass

    class MyDataset(Dataset[MyModel]):
        pass

    This can also be done in a more deeply nested structure, e.g.:

    class MyNumberList(Model[List[int]]):
        pass

    class MyToplevelDict(Model[Dict[str, MyNumberList]]):
        pass

    class MyDataset(Dataset[MyToplevelDict]):
        pass

    Note: the naming of the classes in the examples are for illustrative purposes only and should
    not read as a naming standard.
    """
    class Config:
        validate_assignment = True
        # json_loads = orjson.loads
        # json_dumps = orjson_dumps

    data: Dict[str, ModelT] = Field(default={})

    def __class_getitem__(cls, model: ModelT) -> ModelT:
        # TODO: change model type to params: Union[Type[Any], Tuple[Type[Any], ...]]
        #       as in GenericModel

        # For now, only singular model types are allowed. These lines are needed for
        # interoperability with pydantic GenericModel, which internally stores the model
        # as a tuple:
        if isinstance(model, tuple) and len(model) == 1:
            model = model[0]

        if not isinstance(model, TypeVar) and not lenient_issubclass(model, Model):
            raise TypeError('Invalid model: {}! '.format(model)
                            + 'omnipy Dataset models must be a specialization of the omnipy '
                            'Model class.')
        return super().__class_getitem__(model)

    def __init__(self,
                 value: Union[Dict[str, Any], Iterator[Tuple[str, Any]]] = Undefined,
                 **input_data: Any) -> None:
        if value != Undefined:
            input_data[DATA_KEY] = value

        if self.get_model_class() == ModelT:
            self._raise_no_model_exception()

        GenericModel.__init__(self, **input_data)
        UserDict.__init__(self, self.data)  # noqa
        if not self.__doc__:
            self._set_standard_field_description()

    # TODO: Add test for get_model_class

    def get_model_class(self) -> ModelT:
        return self.__fields__.get(DATA_KEY).type_

    # TODO: Update _raise_no_model_exception() text. Model is now a requirement

    @staticmethod
    def _raise_no_model_exception() -> None:
        raise TypeError(
            'Note: The Dataset class requires a concrete model to be specified as '
            'a type hierarchy within brackets either directly, e.g.:\n\n'
            '\tmodel = Dataset[List[int]]()\n\n'
            'or indirectly in a subclass definition, e.g.:\n\n'
            '\tclass MyNumberListDataset(Dataset[List[int]]): ...\n\n'
            'In both cases, the use of the Model class or a subclass is encouraged if anything '
            'other than the simplest cases, e.g.:\n\n'
            '\tclass MyNumberListModel(Model[List[int]]): ...\n'
            '\tclass MyDataset(Dataset[MyNumberListModel]): ...\n\n'
            'Usage of Dataset without a type specification results in this exception. '
            'Similar use of the Model class do not currently result in an exception, only '
            'a warning message the first time this is done. However, this is just a '
            '"poor man\'s exception" due to complex technicalities in that class. Please '
            'explicitly specify types in both cases. ')

    def _set_standard_field_description(self) -> None:
        self.__fields__[DATA_KEY].field_info.description = self._get_standard_field_description()

    @classmethod
    def _get_standard_field_description(cls) -> str:
        return ('This class represents a data in the `omnipy` Python package and contains '
                'a set of named data items that follows the same data model. '
                'It is a statically typed specialization of the Dataset class according to a '
                'particular specialization of the Model class. Both main classes are wrapping '
                'the excellent Python package named `pydantic`.')

    def __setitem__(self, obj_type: str, data_obj: Any) -> None:
        has_prev_value = obj_type in self.data
        prev_value = self.data.get(obj_type)

        try:
            self.data[obj_type] = data_obj
            self._validate(obj_type)
        except:  # noqa
            if has_prev_value:
                self.data[obj_type] = prev_value
            else:
                del self.data[obj_type]
            raise

    def __getitem__(self, obj_type: str) -> Any:
        if obj_type in self.data:
            return self.data[obj_type].contents
        else:
            return self.data[obj_type]

    def _validate(self, _obj_type: str) -> None:
        self.data = self.data  # Triggers pydantic validation, as validate_assignment=True

    def __iter__(self) -> Iterator:
        return UserDict.__iter__(self)

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in self.__dict__ or attr == DATA_KEY or attr.startswith('__'):
            super().__setattr__(attr, value)
        else:
            raise RuntimeError('Model does not allow setting of extra attributes')

    def to_data(self) -> Dict[str, Any]:
        return GenericModel.dict(self).get(DATA_KEY)

    def from_data(self,
                  data: Union[Dict[str, Any], Iterator[Tuple[str, Any]]],
                  update: bool = True) -> None:
        if not isinstance(data, dict):
            data = dict(data)

        if not update:
            self.clear()

        for obj_type, obj_val in data.items():
            new_model = self.get_model_class()()  # noqa
            new_model.from_data(obj_val)
            self[obj_type] = new_model

    def to_json(self, pretty=False) -> Dict[str, str]:
        result = {}

        for key, val in self.to_data().items():
            result[key] = self._pretty_print_json(val) if pretty else json.dumps(val)
        return result

    def from_json(self,
                  data: Union[Dict[str, str], Iterator[Tuple[str, str]]],
                  update: bool = True) -> None:
        if not isinstance(data, dict):
            data = dict(data)

        if not update:
            self.clear()

        for obj_type, obj_val in data.items():
            new_model = self.get_model_class()()  # noqa
            new_model.from_json(obj_val)
            self[obj_type] = new_model

    # @classmethod
    # def get_type_args(cls):
    #     return cls.__fields__.get(DATA_KEY).type_
    #
    #
    # @classmethod
    # def create_from_json(cls, data: Union[str, Tuple[str]]):
    #     if isinstance(data, tuple):
    #         data = data[0]
    #
    #     obj = cls()
    #     obj.from_json(data, update=False)
    #     return obj
    #
    # def __reduce__(self):
    #     return self.__class__.create_from_json, (self.to_json(),)

    @classmethod
    def to_json_schema(cls, pretty=False) -> Union[str, Dict[str, str]]:
        result = {}
        schema = cls.schema()
        for key, val in schema['properties']['data'].items():
            result[key] = val
        result['title'] = schema['title']
        result['definitions'] = schema['definitions']
        if pretty:
            return cls._pretty_print_json(result)
        else:
            return json.dumps(result)

    @staticmethod
    def _pretty_print_json(json_content: Any) -> str:
        return json.dumps(json_content, indent=4)

    def as_multi_model_dataset(self) -> MultiModelDataset[ModelT]:
        multi_model_dataset = MultiModelDataset[self.get_model_class()]()
        for obj_type in self:
            multi_model_dataset.data[obj_type] = self.data[obj_type]
        return multi_model_dataset


# TODO: Use json serializer package from the pydantic config instead of 'json'


class MultiModelDataset(Dataset[ModelT], Generic[ModelT]):

    _custom_field_models: Dict[str, ModelT] = PrivateAttr(default={})

    def set_model(self, obj_type: str, model: ModelT) -> None:
        try:
            self._custom_field_models[obj_type] = model
            if obj_type in self.data:
                self._validate(obj_type)
            else:
                self.data[obj_type] = model()
        except ValidationError:
            del self._custom_field_models[obj_type]
            raise

    def get_model(self, obj_type: str) -> ModelT:
        if obj_type in self._custom_field_models:
            return self._custom_field_models[obj_type]
        else:
            return self.get_model_class()

    def _validate(self, obj_type: str) -> None:
        if obj_type in self._custom_field_models:
            model = self._custom_field_models[obj_type]
            if not isinstance(model, Model):
                model = Model[model]
            data_obj = self._to_data_if_model(self.data[obj_type])
            parsed_data = self._to_data_if_model(model(data_obj))
            self.data[obj_type] = parsed_data
        super()._validate(obj_type)  # validates all data according to ModelT

    @staticmethod
    def _to_data_if_model(data_obj: Any):
        if isinstance(data_obj, Model):
            data_obj = data_obj.to_data()
        return data_obj



from typing import Any, Dict


class ParamKeyMapper:
    def __init__(self, key_map: Dict[str, str]):
        self.key_map = dict(key_map)
        self._inverse_key_map = {val: key for key, val in self.key_map.items()}

        if len(self.key_map) != len(self._inverse_key_map):
            raise ValueError('Some values were dropped when translating to the inverse key_map!')

    def _get_key_map(self, inverse: bool):
        return self._inverse_key_map if inverse else self.key_map

    def map_matching_keys(self, params: Dict[str, Any], inverse: bool,
                          keep_non_matching_keys: bool):
        key_map = self._get_key_map(inverse)
        keys = params.keys() if keep_non_matching_keys else [
            key for key in params if key in key_map
        ]
        return {key_map[key] if key in key_map else key: params[key] for key in keys}

    def delete_matching_keys(self, params: Dict[str, Any], inverse: bool):
        key_map = self._get_key_map(inverse)
        return {key: params[key] for key in params if key not in key_map}

    def map_matching_keys_delete_inverse_matches_keep_rest(self,
                                                           params: Dict[str, Any],
                                                           inverse: bool):
        params = self.delete_matching_keys(params, not inverse)
        params = self.map_matching_keys(params, inverse, keep_non_matching_keys=True)

        return params



__version__ = '0.7.0'

import os
import sys

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

runtime = None

if 'pytest' not in sys.modules:
    from omnipy.hub.runtime import Runtime
    runtime = Runtime()



from collections import defaultdict
from copy import copy
from functools import lru_cache
from typing import Dict, List, Tuple

import jq

from omnipy.compute.task import TaskTemplate
from omnipy.data.dataset import Dataset

from ..json.models import (JsonDataset,
                           JsonDictOfAnyModel,
                           JsonDictOfListOfDictOfAnyModel,
                           JsonListOfDictOfAnyModel,
                           JsonListOfNestedDictsModel)
from ..pandas.models import PandasDataset
from .models import TableOfStrings, TableOfStringsAndLists

ID_KEY = '_omnipy_id'
REF_KEY = '_omnipy_ref'
DEFAULT_KEY = '__root__'


@TaskTemplate()
def remove_columns(
        json_dataset: JsonDataset,
        column_keys_for_data_items: Dict[str, List[str]]) -> Dataset[TableOfStringsAndLists]:
    # TODO: implement general solution to make sure that one does not modify input data by
    #       automatically copying or otherwise
    #
    # dataset_copy = deepcopy(json_dataset)
    for data_item_key, column_keys in column_keys_for_data_items.items():
        for record in json_dataset[data_item_key]:
            for column in column_keys:
                if column in record:
                    del record[column]
    return Dataset[TableOfStringsAndLists](json_dataset.to_data())


@TaskTemplate()
def convert_to_1nf(input_dataset: Dataset[TableOfStringsAndLists]) -> Dataset[TableOfStrings]:
    out_dataset = Dataset[TableOfStrings]()
    for table_name, table in input_dataset.items():
        # for record in table:
        #     ...
        out_dataset[table_name] = table
    return out_dataset


@lru_cache(maxsize=1)
def get_filter_all_top_level_parents_of_subtrees_compiler():
    return jq.compile('map_values(arrays,objects)')


def filter_all_top_level_parents_of_subtrees(data: JsonDictOfAnyModel) -> JsonDictOfAnyModel:
    return get_filter_all_top_level_parents_of_subtrees_compiler().input(data).first()


def delete_element_from_dict(data, path):
    path_list = path.split('.')
    for i in path_list[:-1]:
        data = data[i]
    del data[path_list[-1]]


def update_dict_from_front(dict_a: Dict, dict_b: Dict) -> None:
    dict_a_copy = copy(dict_a)
    dict_a.clear()
    dict_a.update(dict_b)
    dict_a.update(dict_a_copy)


def flatten_outer_level_of_nested_record(nested_record: JsonDictOfAnyModel,
                                         record_id: str,
                                         data_file_title: str,
                                         id_key: str,
                                         ref_key: str,
                                         default_key: str) -> JsonDictOfListOfDictOfAnyModel:
    all_top_level_parents_of_subtrees = filter_all_top_level_parents_of_subtrees(nested_record)
    record_id = ensure_item_first_in_nested_record(
        nested_record, key=id_key, default_value=record_id, update=False)

    new_data_files = {}
    for key, value in all_top_level_parents_of_subtrees.items():
        new_data_file_title: str = f'{data_file_title}.{key}'
        new_data_file = value

        if isinstance(new_data_file, dict):
            new_data_file = [new_data_file]

        add_reference_to_parent_from_child(
            child=new_data_file,
            parent_title=data_file_title,
            ident=record_id,
            ref_key=ref_key,
            default_key=default_key)

        new_data_files[new_data_file_title] = new_data_file
        delete_element_from_dict(nested_record, key)

    return new_data_files


def get_ref_value(data_file_title, ident):
    return f'{data_file_title}.{ident}'


def add_reference_to_parent_from_child(
    child: JsonListOfDictOfAnyModel,
    parent_title: str,
    ident: str,
    ref_key: str,
    default_key: str,
):
    ref_value = get_ref_value(parent_title, ident)
    for i, new_nested_record in enumerate(child):
        if not isinstance(new_nested_record, dict):
            new_nested_record = {default_key: new_nested_record}
            child[i] = new_nested_record
        ensure_item_first_in_nested_record(
            new_nested_record, key=ref_key, default_value=ref_value, fail_if_present=True)


def ensure_item_first_in_nested_record(
    nested_record,
    key,
    default_value,
    update=True,
    fail_if_present=False,
):
    value = default_value

    if key in nested_record:
        assert not fail_if_present, f'Key "{key}" already present in dict'
        if not update:
            value = nested_record[key]
        del nested_record[key]

    update_dict_from_front(nested_record, {key: value})

    return value


def split_outer_lists_from_dataset_as_new_data_files(
    dataset: Dataset[JsonListOfDictOfAnyModel], id_key: str, ref_key: str, default_key: str
) -> Tuple[Dataset[JsonListOfNestedDictsModel], Dataset[JsonListOfDictOfAnyModel]]:

    data_files_without_nested_lists = defaultdict(list)
    data_files_maybe_with_nested_lists = defaultdict(list)
    for data_file_title, item in dataset.items():
        data_file = item.to_data()

        if len(data_file) == 0:
            data_files_without_nested_lists[data_file_title] = []

        for record_id, nested_record in enumerate(data_file):
            new_data_files = flatten_outer_level_of_nested_record(
                nested_record,
                str(record_id),
                data_file_title,
                id_key,
                ref_key,
                default_key,
            )

            for new_data_file_title, new_data_file in new_data_files.items():
                data_files_maybe_with_nested_lists[new_data_file_title] += new_data_file
            data_files_without_nested_lists[data_file_title].append(nested_record)

    all_data_files_without_nested_lists_ds = Dataset[JsonListOfNestedDictsModel](
        data_files_without_nested_lists)
    all_data_files_maybe_with_nested_lists_ds = Dataset[JsonListOfDictOfAnyModel](
        data_files_maybe_with_nested_lists)

    return all_data_files_without_nested_lists_ds, all_data_files_maybe_with_nested_lists_ds


@TaskTemplate(fixed_params=dict(id_key=ID_KEY, ref_key=REF_KEY, default_key=DEFAULT_KEY))
def flatten_nested_json_to_list_of_dicts(
    dataset: Dataset[JsonListOfDictOfAnyModel],
    id_key: str,
    ref_key: str,
    default_key: str,
) -> Dataset[JsonListOfNestedDictsModel]:
    data_files_maybe_with_nested_lists_ds = dataset
    all_data_files_without_nested_lists_ds = Dataset[JsonListOfNestedDictsModel]()

    while len(data_files_maybe_with_nested_lists_ds) > 0:
        data_files_without_nested_lists_ds, data_files_maybe_with_nested_lists_ds = \
            split_outer_lists_from_dataset_as_new_data_files(
                data_files_maybe_with_nested_lists_ds, id_key, ref_key, default_key)
        all_data_files_without_nested_lists_ds.update(data_files_without_nested_lists_ds)

    return all_data_files_without_nested_lists_ds


@TaskTemplate()
def transpose_dataset_of_dicts_to_lists(dataset: Dataset[JsonDictOfAnyModel],
                                        id_key: str = ID_KEY) -> Dataset[JsonListOfDictOfAnyModel]:
    output_dataset = Dataset[JsonListOfDictOfAnyModel]()
    output_data = defaultdict(list)
    for name, item in dataset.items():
        for key, val in item.to_data().items():
            if not isinstance(val, list):
                val = [val]

            for item_index, item in enumerate(val):
                if isinstance(item, dict):
                    data = {id_key: f'{name}_{item_index}'}
                    assert id_key not in item
                    data.update(item)
                else:
                    data = item
                output_data[key].append(data)
    output_dataset.update(output_data.items())
    return output_dataset



from typing import Optional


class ResultKeyFuncJobBaseMixin:
    # Requires NameJobBaseMixin

    def __init__(self, *, result_key: Optional[str] = None):
        self._result_key = result_key

        if self.result_key is not None:
            self._check_not_empty_string('result_key', self.result_key)

    @property
    def result_key(self) -> Optional[str]:
        return self._result_key

    def _call_job(self, *args: object, **kwargs: object) -> object:

        result = super()._call_job(*args, **kwargs)

        if self._result_key:
            return {self._result_key: result}
        else:
            return result



import inspect
from types import MappingProxyType
from typing import Dict, Iterable, Type


class SignatureFuncJobBaseMixin:
    def __init__(self):
        self._func_signature = inspect.signature(self._job_func)
        self.__signature__ = self._func_signature

    @property
    def param_signatures(self) -> MappingProxyType:
        return self._func_signature.parameters

    @property
    def return_type(self) -> Type[object]:
        return self._func_signature.return_annotation

    def _check_param_keys_in_func_signature(self,
                                            param_keys: Iterable[str],
                                            modifier_kwarg_key: str) -> None:
        if any(param.kind == inspect.Parameter.VAR_KEYWORD
               for param in self.param_signatures.values()):
            return

        for param_key in param_keys:
            if param_key not in self.param_signatures:
                raise KeyError('Parameter "{}" was not found in the '.format(param_key)
                               + 'signature of the job function. Only parameters in the '
                               'signature of the job function are '
                               f'allowed as keys in the "{modifier_kwarg_key}" modifier: '
                               f"{', '.join(key for key in self.param_signatures.keys())}")

    def _update_func_signature(self, new_signature: inspect.Signature):
        # self._job_func.__signature__ = new_signature
        self._func_signature = new_signature
        self.__signature__ = new_signature

    def get_call_args(self, *args: object, **kwargs: object) -> Dict[str, object]:
        return self._func_signature.bind(*args, **kwargs).arguments



from typing import Callable, Tuple

from omnipy.compute.func_job import FuncArgJobBase
from omnipy.compute.task import TaskTemplate

# class FlowInit(FuncArgJobBase):
#     def __init__(
#         self,
#         job_func: Callable,
#         *args: object,
#         name: Optional[str] = None,
#         iterate_over_data_files: bool = False,
#         fixed_params: Optional[Mapping[str, object]] = None,
#         param_key_map: Optional[Mapping[str, str]] = None,
#         result_key: Optional[str] = None,
#         persist_outputs: Optional[PersistOutputsOptions] = None,
#         restore_outputs: Optional[RestoreOutputsOptions] = None,
#         **kwargs: object,
#     ):
#         # super().__init__(
#         #     job_func,
#         #     **remove_none_vals(
#         #         name=name,
#         #         iterate_over_data_files=iterate_over_data_files,
#         #         fixed_params=fixed_params,
#         #         param_key_map=param_key_map,
#         #         result_key=result_key,
#         #         persist_outputs=persist_outputs,
#         #         restore_outputs=restore_outputs,
#         #         **kwargs,
#         #     ))
#
#         self._time_of_last_run = None

# class FlowTemplate(FuncJobTemplate, ABC):
#     ...

# class TaskTemplatesFlowBase(FlowInit):
#     def __init__(
#         self,
#         job_func: Callable,
#         *task_templates: TaskTemplate,
#         name: Optional[str] = None,
#         iterate_over_data_files: bool = False,
#         fixed_params: Optional[Mapping[str, object]] = None,
#         param_key_map: Optional[Mapping[str, str]] = None,
#         result_key: Optional[str] = None,
#         persist_outputs: Optional[PersistOutputsOptions] = None,
#         restore_outputs: Optional[RestoreOutputsOptions] = None,
#         **kwargs: object,
#     ):
#         # super().__init__(
#         #     job_func,
#         #     **remove_none_vals(
#         #         name=name,
#         #         iterate_over_data_files=iterate_over_data_files,
#         #         fixed_params=fixed_params,
#         #         param_key_map=param_key_map,
#         #         result_key=result_key,
#         #         persist_outputs=persist_outputs,
#         #         restore_outputs=restore_outputs,
#         #         **kwargs,
#         #     ))


class TaskTemplateArgsJobBase(FuncArgJobBase):
    def __init__(self, job_func: Callable, *task_templates: TaskTemplate, **kwargs: object) -> None:
        # FuncArgJobBase.__init__(self, job_func, *task_templates, **kwargs)

        self._task_templates: Tuple[TaskTemplate, ...] = task_templates

    def _get_init_args(self) -> Tuple[object, ...]:
        return self._job_func, *self._task_templates

    @property
    def task_templates(self) -> Tuple[TaskTemplate, ...]:
        return self._task_templates

    def _refine(self,
                *task_templates: TaskTemplate,
                update: bool = True,
                **kwargs: object):  # -> TaskTemplatesFlowTemplateT:

        return super()._refine(
            *task_templates,
            update=update,
            **kwargs,
        )


#
# class TaskTemplatesFlowTemplate(TaskTemplatesFlowBase, ABC):
#     def refine(self,
#                *task_templates: TaskTemplate,
#                update: bool = True,
#                name: Optional[str] = None,
#                iterate_over_data_files: bool = False,
#                fixed_params: Optional[Mapping[str, object]] = None,
#                param_key_map: Optional[Mapping[str, str]] = None,
#                result_key: Optional[str] = None,
#                persist_outputs: Optional[PersistOutputsOptions] = None,
#                restore_outputs: Optional[RestoreOutputsOptions] = None,
#                **kwargs: object):  # -> TaskTemplatesFlowTemplateT:
#
#         return self._refine(
#             self,
#             *task_templates,
#             update=update,
#             **remove_none_vals(
#                 name=name,
#                 iterate_over_data_files=iterate_over_data_files,
#                 fixed_params=fixed_params,
#                 param_key_map=param_key_map,
#                 result_key=result_key,
#                 persist_outputs=persist_outputs,
#                 restore_outputs=restore_outputs,
#                 **kwargs,
#             ))
#
#
# class TaskTemplatesFlow(TaskTemplatesFlowBase, Flow, ABC):
#     ...



from datetime import datetime
import os

from omnipy import ROOT_DIR
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

from .serializers import RawDatasetToTarFileSerializer


def serialize_to_tarpacked_raw_files(dataset_name: str, dataset: Dataset[Model[str]]) -> None:
    """
    Serializing is not fully implemented, so here is a helper method to write a dataset to a
    dataset_name.tar.gz file with a set of csv files, one for each data item (endpoint/table)
    """
    data_dir = os.path.join(ROOT_DIR,
                            '..',
                            '..',
                            'data',
                            datetime.now().strftime('%Y_%m_%d-%H_%M_%S'))

    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    file_path = os.path.join(data_dir, dataset_name + '.tar.gz')

    print('Writing dataset as a gzipped tarpack of raw files to "{}"'.format(
        os.path.abspath(file_path)))
    serializer = RawDatasetToTarFileSerializer()
    with open(file_path, 'wb') as tarfile:
        tarfile.write(serializer.serialize(dataset))



from typing import List, Protocol


class IsModifyContentsCallable(Protocol):
    def __call__(self, data_file: str, **kwargs: object) -> str:
        ...


class IsModifyEachLineCallable(Protocol):
    def __call__(self, line_no: int, line: str, **kwargs: object) -> str:
        ...


class IsModifyAllLinesCallable(Protocol):
    def __call__(self, all_lines: List[str], **kwargs: object) -> List[str]:
        ...



from typing import Optional

from inflection import underscore
from prefect.utilities.names import generate_slug
from slugify import slugify

from omnipy.util.mixin import DynamicMixinAcceptor


class NameJobBaseMixin:
    def __init__(self, *, name: Optional[str] = None):
        self._name: Optional[str] = name

        if self._name is not None:
            self._check_not_empty_string('name', self._name)
        else:
            if hasattr(self, '_job_func'):
                self._name = self._job_func.__name__

        # TODO: When job state machine is implemented, check using that to see if in job state
        self._unique_name = self._generate_unique_name() if hasattr(self, 'create_job') else None

    @staticmethod
    def _check_not_empty_string(param_name: str, param: str) -> None:
        if len(param) == 0:
            raise ValueError('Empty strings not allowed for parameter "{}"'.format(param_name))

    @property
    def name(self) -> str:
        return self._name

    @property
    def unique_name(self) -> str:
        return self._unique_name

    def _generate_unique_name(self) -> str:
        if self._name is None:
            return None

        class_name = self.__class__.__name__
        if class_name.endswith(DynamicMixinAcceptor.WITH_MIXINS_CLS_PREFIX):
            class_name = class_name[:-len(DynamicMixinAcceptor.WITH_MIXINS_CLS_PREFIX)]

        class_name_snake_case = underscore(class_name)
        return slugify(f'{class_name_snake_case}-{self._name}-{generate_slug(2)}')

    def _regenerate_unique_name(self) -> None:
        self._unique_name = self._generate_unique_name()


class NameJobMixin:
    def regenerate_unique_name(self) -> None:
        self._regenerate_unique_name()


#
# class NameFuncJobBaseMixin:
#     def __init__(self, *, name: Optional[str] = None):
#         self._name = name if name is not None else self._job_func.__name__



from types import MappingProxyType
from typing import Mapping, Optional

from omnipy.util.helpers import repr_max_len
from omnipy.util.param_key_mapper import ParamKeyMapper


class ParamsFuncJobBaseMixin:
    # Requires FuncSignatureJobBaseMixin

    def __init__(
        self,
        *,
        fixed_params: Optional[Mapping[str, object]] = None,
        param_key_map: Optional[Mapping[str, str]] = None,
    ):
        self._fixed_params = dict(fixed_params) if fixed_params is not None else {}
        self._param_key_mapper = ParamKeyMapper(param_key_map if param_key_map is not None else {})

        self._check_param_keys_in_func_signature(self.fixed_params.keys(), 'fixed_params')
        self._check_param_keys_in_func_signature(self.param_key_map.keys(), 'param_key_map')

    @property
    def fixed_params(self) -> MappingProxyType[str, object]:
        return MappingProxyType(self._fixed_params)

    @property
    def param_key_map(self) -> MappingProxyType[str, str]:
        return MappingProxyType(self._param_key_mapper.key_map)

    def _call_job(self, *args: object, **kwargs: object) -> object:

        mapped_fixed_params = self._param_key_mapper.delete_matching_keys(
            self._fixed_params, inverse=True)
        mapped_kwargs = self._param_key_mapper.map_matching_keys_delete_inverse_matches_keep_rest(
            kwargs, inverse=True)
        try:
            if len(mapped_kwargs) < len(kwargs):
                raise TypeError('Keyword arguments {} matches parameter key map inversely.'.format(
                    tuple(set(kwargs.keys()) - set(mapped_kwargs.keys()))))

            result = super()._call_job(*args, **mapped_fixed_params, **mapped_kwargs)

        except TypeError as e:
            if not str(e).startswith('Incorrect job function arguments'):
                raise TypeError(
                    f'Incorrect job function arguments for job "{self.name}"!\n'
                    f'Job class name: {self.__class__.__name__}\n'
                    f'Current parameter key map contents: {self.param_key_map}\n'
                    f'Positional arguments: {repr_max_len(args)}\n'
                    f'Keyword arguments: {repr_max_len(kwargs)}\n'
                    f'Mapped fixed parameters: {repr_max_len(mapped_fixed_params)}\n'
                    f'Mapped keyword arguments: {repr_max_len(mapped_kwargs)}\n'
                    f'Call function signature parameters: '
                    f'{[(str(p), p.kind) for p in self.param_signatures.values()]}') from e
            else:
                raise

        return result



import requests

ENCODE_HEADERS = {'accept': 'application/json'}
ENCODE_BASE_URL = 'https://www.encodeproject.org/'
GDC_BASE_URL = 'https://api.gdc.cancer.gov/'


# ['experiments', 'biosample']
def encode_api(cls, endpoint='experiments', id=None, limit=None, format='json', frame='object'):
    api_url = ENCODE_BASE_URL + endpoint + '/' + (id if id else '@@listing') + '?' + '&'.join(
        (['limit=' + limit] if limit else []) + (['format=' + format] if format else [])
        + (['frame=' + frame] if frame else []))
    print(api_url)
    response = requests.get(api_url, headers=ENCODE_HEADERS)
    if response.status_code == 200:
        results = response.json()
        if results['notification'] == 'Success':
            graph = results['@graph']
            return graph
    else:
        print('No result found')


# ['projects', 'cases', 'files', 'annotations'], starting_point='0', size='25'
def gdc_api(cls, object_type='projects', starting_point=None, size=None):
    api_url = GDC_BASE_URL + object_type + '/' + '?' + \
              '&'.join(
                  (['from=' + starting_point] if starting_point else [])
                  + (['size=' + size] if size else [])
                  + (['expand=' + 'project'] if object_type == 'cases' else [])
              )
    print(api_url)
    response = requests.get(api_url)
    if response.status_code != 200:
        print('No result found')
        return None

    results = response.json()

    if len(results['warnings']) > 0:
        print('The following warnings have been encountered:')
        print(results['warnings'])
        return None

    hits = (results['data']['hits'])
    return hits



import os
from typing import Callable, cast, List, Tuple, TypeVar

from omnipy.compute.task import TaskTemplate
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

ModelT = TypeVar('ModelT', bound=Model)


@TaskTemplate()
def cast_dataset(dataset: Dataset, cast_model: Callable[[], ModelT]) -> ModelT:
    out_dataset: Dataset[ModelT] = Dataset[cast_model]()
    for name, item in dataset.items():
        out_dataset[name] = cast(cast_model, item.to_data())
    return out_dataset


@TaskTemplate()
def split_dataset(
        dataset: Dataset[Model[object]],
        datafile_names_for_b: List[str]) -> Tuple[Dataset[Model[object]], Dataset[Model[object]]]:
    model_cls = dataset.get_model_class()
    datafile_names_for_a = set(dataset.keys()) - set(datafile_names_for_b)
    dataset_a = Dataset[model_cls](
        {name: dataset[name] for name in dataset.keys() if name in datafile_names_for_a})
    dataset_b = Dataset[model_cls](
        {name: dataset[name] for name in dataset.keys() if name in datafile_names_for_b})
    return dataset_a, dataset_b


@TaskTemplate()
def import_directory(directory: str,
                     suffix: str = '.json',
                     model: Model = Model[str]) -> Dataset[Model]:
    dataset = Dataset[model]()
    for import_filename in os.listdir(directory):
        if import_filename.endswith(suffix):
            with open(os.path.join(directory, import_filename)) as open_file:
                dataset_name = '_'.join(import_filename.split('.')[:-1])
                print(f"{import_filename} -> Dataset['{dataset_name}']")
                dataset[dataset_name] = open_file.read()
    return dataset



from typing import Any, Callable, Type

from omnipy.api.protocols import IsDagFlow, IsFuncFlow, IsLinearFlow, IsLocalRunnerConfig, IsTask
from omnipy.config.engine import LocalRunnerConfig
from omnipy.engine.job_runner import (DagFlowRunnerEngine,
                                      FuncFlowRunnerEngine,
                                      LinearFlowRunnerEngine,
                                      TaskRunnerEngine)


class LocalRunner(TaskRunnerEngine,
                  LinearFlowRunnerEngine,
                  DagFlowRunnerEngine,
                  FuncFlowRunnerEngine):
    def _init_engine(self) -> None:
        ...

    def _update_from_config(self) -> None:
        ...

    @classmethod
    def get_config_cls(cls) -> Type[IsLocalRunnerConfig]:
        return LocalRunnerConfig

    def _init_task(self, task: IsTask, call_func: Callable) -> Any:
        ...

    def _run_task(self, state: Any, task: IsTask, call_func: Callable, *args, **kwargs) -> Any:
        return call_func(*args, **kwargs)

    def _init_linear_flow(self, flow: IsLinearFlow) -> Any:
        ...

    def _run_linear_flow(self, state: Any, flow: IsLinearFlow, *args, **kwargs) -> Any:
        return self.default_linear_flow_run_decorator(flow)(*args, **kwargs)

    def _init_dag_flow(self, flow: IsDagFlow) -> Any:
        ...

    def _run_dag_flow(self, state: Any, flow: IsDagFlow, *args, **kwargs) -> Any:
        return self.default_dag_flow_run_decorator(flow)(*args, **kwargs)

    def _init_func_flow(self, func_flow: IsFuncFlow, call_func: Callable) -> object:
        pass

    def _run_func_flow(self,
                       state: Any,
                       func_flow: IsFuncFlow,
                       call_func: Callable,
                       *args,
                       **kwargs) -> Any:
        with func_flow.flow_context:
            return call_func(*args, **kwargs)



# from __future__ import annotations
from typing import Dict, ForwardRef, Generic, List, TypeVar, Union

from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

JsonType = ForwardRef('JsonType')

JsonListValT = TypeVar('JsonListValT')
JsonDictValT = TypeVar('JsonDictValT')

# class JsonList(GenericModel, Generic[JsonListValT]):
#     __root__: List[JsonListValT] = []
#
#     @property
#     def contents(self):
#         return self.__root__
#
#     def to_data(self) -> Any:
#         return self.dict()[ROOT_KEY]
#
#
# class JsonDict(GenericModel, Generic[JsonDictValT]):
#     __root__: Dict[str, JsonDictValT] = {}
#
#     @property
#     def contents(self):
#         return self.__root__
#
#     def to_data(self) -> Any:
#         return self.dict()[ROOT_KEY]


class JsonList(Model[List[Union[None, JsonListValT]]], Generic[JsonListValT]):
    ...


class JsonDict(Model[Dict[str, Union[None, JsonDictValT]]], Generic[JsonDictValT]):
    ...


class JsonScalarType(Model[Union[None, int, float, str, bool]]):
    ...


JsonType = Union[None, JsonScalarType, JsonList['JsonType'], JsonDict['JsonType']]

JsonList['JsonType'].update_forward_refs(JsonType=JsonType)
JsonDict['JsonType'].update_forward_refs(JsonType=JsonType)


class JsonModel(Model[JsonType]):
    ...


class JsonDictOfAnyModel(Model[JsonDict[JsonType]]):
    ...


class JsonDictOfDictOfAnyModel(Model[JsonDict[JsonDict[JsonType]]]):
    ...


class JsonListOfAnyModel(Model[Union[JsonList[JsonType]]]):
    ...


class JsonListOfDictOfAnyModel(Model[JsonList[JsonDict[JsonType]]]):
    ...


class JsonDictOfListOfDictOfAnyModel(Model[JsonDict[JsonListOfDictOfAnyModel]]):
    ...


# @classmethod
# def _parse_data(cls, data: Union[JsonList, JsonDict]) -> Union[JsonList, JsonDict]:
#     # data = cls._data_not_empty_object(data)
#     return data

# @classmethod
# def _data_not_empty_object(cls, data: List):
#     for obj in data:
#         assert len(obj) > 0
#     return data


class JsonDataset(Dataset[JsonModel]):
    ...


JsonNoListType = Union[JsonScalarType, JsonDict['JsonType']]

JsonList['JsonNoListType'].update_forward_refs(JsonNoListType=JsonNoListType)
JsonDict['JsonNoListType'].update_forward_refs(JsonNoListType=JsonNoListType)


class JsonListOfDictsModel(Model[JsonList[JsonDict[JsonScalarType]]]):
    ...


class JsonNestedDictsModel(Model[JsonDict[JsonNoListType]]):
    ...


class JsonListOfNestedDictsModel(Model[JsonList[JsonNestedDictsModel]]):
    ...



from collections import defaultdict
import inspect
import types
from typing import DefaultDict, Dict, List, Protocol, Type

from omnipy.util.helpers import generic_aware_issubclass, get_bases, transfer_generic_args_to_cls


class IsMixin(Protocol):
    def __init__(self, **kwargs: object) -> None:
        ...


class DynamicMixinAcceptor:
    # Constants
    WITH_MIXINS_CLS_PREFIX = 'WithMixins'

    # Declarations needed by mypy
    _orig_class: Type
    _orig_init_signature: inspect.Signature
    _mixin_classes: List[Type]
    _init_params_per_mixin_cls: DefaultDict[str, DefaultDict[str, inspect.Parameter]]

    def __class_getitem__(cls, item):
        return super().__class_getitem__(item)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        if DynamicMixinAcceptor in get_bases(cls) and cls.__init__ is object.__init__:
            raise TypeError('Mixin acceptor class is required to define a __init__() method.')

        cls._orig_init_signature = inspect.signature(cls.__init__)
        cls._mixin_classes = []
        cls._init_params_per_mixin_cls = defaultdict(defaultdict)

    @classmethod
    def _get_mixin_init_kwarg_params(cls) -> Dict[str, inspect.Parameter]:
        return {
            key: param for param_dict in cls._init_params_per_mixin_cls.values() for key,
            param in param_dict.items()
        }

    @property
    def _mixin_init_kwarg_params(self) -> Dict[str, inspect.Parameter]:
        return self._get_mixin_init_kwarg_params()

    @classmethod
    def _get_mixin_init_kwarg_params_including_bases(cls) -> Dict[str, inspect.Parameter]:
        all_mixin_init_kwarg_params: Dict[str, inspect.Parameter] = {}

        base_list = list(cls.__mro__)
        skip_bases = {'DynamicMixinAcceptor'}

        for base in base_list:
            if base.__name__.endswith(cls.WITH_MIXINS_CLS_PREFIX):
                skip_bases.add(base.__name__[:-len(cls.WITH_MIXINS_CLS_PREFIX)])

        cleaned_base_list = [
            base for base in base_list
            if issubclass(base, DynamicMixinAcceptor) and base.__name__ not in skip_bases
        ]

        for base in cleaned_base_list:
            all_mixin_init_kwarg_params |= dict(base._get_mixin_init_kwarg_params().items())

        return all_mixin_init_kwarg_params

    @property
    def _mixin_init_kwarg_params_including_bases(self) -> Dict[str, inspect.Parameter]:
        return self._get_mixin_init_kwarg_params_including_bases()

    @classmethod
    def accept_mixin(cls, mixin_cls: Type) -> None:
        cls._accept_mixin(mixin_cls, update=True)

    @classmethod
    def _accept_mixin(cls, mixin_cls: Type, update: bool):
        cls._mixin_classes.append(mixin_cls)

        if '__init__' in mixin_cls.__dict__:
            cls._store_init_signature_params_for_mixin(mixin_cls)

            if update:
                cls._update_cls_init_signature_with_kwargs_all_mixin_kwargs()

    @classmethod
    def _store_init_signature_params_for_mixin(cls, mixin_cls):
        for key, param in inspect.signature(mixin_cls.__init__).parameters.items():
            if key != 'self':
                if param.kind not in (param.KEYWORD_ONLY, param.VAR_KEYWORD):
                    raise AttributeError(
                        'All params in the signature of the __init__() method in a dynamic '
                        'mixin class must be keyword-only or var-keyword '
                        '(except for the "self" param)')
                if param.kind == param.KEYWORD_ONLY:
                    cls._init_params_per_mixin_cls[mixin_cls.__name__][key] = param

    @classmethod
    def _update_cls_init_signature_with_kwargs_all_mixin_kwargs(cls):
        updated_init_signature = cls._get_updated_cls_init_signature_with_all_mixin_kwargs()

        cls.__init__.__signature__ = updated_init_signature
        if hasattr(cls, '_orig_class'):
            cls._orig_class.__init____signature__ = updated_init_signature

    @classmethod
    def _get_updated_cls_init_signature_with_all_mixin_kwargs(cls):
        orig_init_param_dict = cls._orig_init_signature.parameters
        init_params = list(orig_init_param_dict.values())
        opt_var_keyword_param = []

        if init_params[-1].kind == inspect.Parameter.VAR_KEYWORD:
            opt_var_keyword_param = [init_params[-1]]
            init_params = init_params[:-1]

        only_mixin_params = [
            val for (key, val) in cls._get_mixin_init_kwarg_params_including_bases().items()
            if key not in orig_init_param_dict
        ]

        updated_params = init_params + only_mixin_params + opt_var_keyword_param
        return cls._orig_init_signature.replace(parameters=updated_params)

    @classmethod
    def reset_mixins(cls):
        cls._mixin_classes.clear()
        cls._init_params_per_mixin_cls.clear()
        cls.__init__.__signature__ = cls._orig_init_signature

    def __new__(cls, *args, **kwargs):
        if not cls.__name__.endswith(cls.WITH_MIXINS_CLS_PREFIX):
            cls_with_mixins = cls._create_subcls_inheriting_from_mixins_and_orig_cls()
            obj = super(cls, cls_with_mixins).__new__(cls_with_mixins, *args, **kwargs)

        else:
            obj = object.__new__(cls)

        cls._update_cls_init_signature_with_kwargs_all_mixin_kwargs()
        return obj

    @classmethod
    def _create_subcls_inheriting_from_mixins_and_orig_cls(cls):

        # TODO: Refactor this, and possibly elsewhere in class

        def __init__(self, *args, **kwargs):
            # print(f'__init__ for obj of class: {self.__class__.__name__}')
            cls.__init__(self, *args, **kwargs)
            mixin_kwargs_defaults = {}

            for base in cls_with_mixins.__mro__:
                if base == cls \
                        or base.__name__.endswith(cls.WITH_MIXINS_CLS_PREFIX) \
                        or base.__init__ is object.__init__:
                    continue

                mixin_kwargs = {}
                contains_positional = False

                for key, param in inspect.signature(base.__init__).parameters.items():
                    if key != 'self':
                        if param.kind == param.KEYWORD_ONLY:
                            if key in kwargs:
                                mixin_kwargs[key] = kwargs[key]
                            elif key in mixin_kwargs_defaults:
                                mixin_kwargs[key] = mixin_kwargs_defaults[key]
                            else:
                                mixin_kwargs_defaults[key] = param.default
                        elif param.kind in (param.POSITIONAL_ONLY,
                                            param.POSITIONAL_OR_KEYWORD,
                                            param.VAR_POSITIONAL):
                            contains_positional = True

                if contains_positional:
                    # print(f'Calling... {base.__name__}(args={args}, kwargs={mixin_kwargs})')
                    base.__init__(self, *args, **mixin_kwargs)
                else:
                    # print(f'Calling... {base.__name__}(kwargs={mixin_kwargs})')
                    base.__init__(self, **mixin_kwargs)

        cls_bases = list(get_bases(cls))
        if not cls.__name__.endswith(cls.WITH_MIXINS_CLS_PREFIX):
            cls_bases = list(cls._mixin_classes) + cls_bases

        cls_bases_with_mixins = []

        for cls_base in cls_bases:

            if cls._is_true_acceptor_subclass(cls_base):
                cls_new_base = cls_base._create_subcls_inheriting_from_mixins_and_orig_cls()
                cls_base = transfer_generic_args_to_cls(cls_new_base, cls_base)

            cls_bases_with_mixins.append(cls_base)

        def fill_ns(ns):
            ns |= dict(__init__=__init__)
            return ns

        cls_with_mixins = types.new_class(
            f'{cls.__name__}{cls.WITH_MIXINS_CLS_PREFIX}',
            tuple([cls] + cls_bases_with_mixins),
            {},
            fill_ns,
        )

        cls_with_mixins._orig_class = cls
        cls_with_mixins._orig_init_signature = inspect.signature(cls.__init__)

        for mixin_cls in cls._mixin_classes:
            cls_with_mixins._accept_mixin(mixin_cls, update=False)

        return cls_with_mixins

    @staticmethod
    def _is_true_acceptor_subclass(cls):
        if cls == DynamicMixinAcceptor:
            return False
        return generic_aware_issubclass(cls, DynamicMixinAcceptor)



import pandas as pd


def split_dataframe(df, row_labels_to_move, col_labels_to_move, row_level=1, col_level=0):
    mv_rows = select_from_labels(df, row_labels_to_move, axis=0, level=row_level)
    kp_rows = select_from_labels(df, row_labels_to_move, axis=0, level=row_level, invert=True)
    mv_cols = select_from_labels(df, col_labels_to_move, axis=1, level=col_level)
    kp_cols = select_from_labels(df, col_labels_to_move, axis=1, level=col_level, invert=True)

    kp_df = extract_subtable_by_labels(df, kp_rows, kp_cols, row_level, col_level)
    new_row_df = extract_subtable_by_labels(df, kp_rows, mv_cols, row_level, col_level)
    new_col_df = extract_subtable_by_labels(df, mv_rows, kp_cols, row_level, col_level)
    new_row_col_df = extract_subtable_by_labels(df, mv_rows, mv_cols, row_level, col_level)

    return kp_df, new_row_df, new_col_df, new_row_col_df


def select_from_labels(df, labels, axis, level, invert=False):
    assert axis in [0, 1]
    assert level in [0, 1]

    idx = df.index.levels[level] if axis == 0 else df.columns.levels[level]
    filtered_idx = idx[~idx.isin(labels)] if invert else idx[idx.isin(labels)]
    formatted_idx = tuple(filtered_idx)
    return formatted_idx


def extract_subtable_by_labels(df,
                               row_labels=None,
                               col_labels=None,
                               row_level=None,
                               col_level=None):
    if row_labels:
        rows = [df.xs(row, axis=0, level=row_level, drop_level=False) for row in row_labels]
        df = pd.concat(rows)
    if col_labels:
        cols = [df.xs(col, axis=1, level=col_level, drop_level=False) for col in col_labels]
        df = pd.concat(cols, axis=1)
    return df



import asyncio
from datetime import timedelta
from typing import Any, Callable, Type

from prefect import flow as prefect_flow
from prefect import Flow as PrefectFlow
from prefect import task as prefect_task
from prefect import Task as PrefectTask
from prefect.tasks import task_input_hash

from omnipy.api.protocols import IsDagFlow, IsFuncFlow, IsLinearFlow, IsPrefectEngineConfig, IsTask
from omnipy.config.engine import PrefectEngineConfig
from omnipy.engine.job_runner import (DagFlowRunnerEngine,
                                      FuncFlowRunnerEngine,
                                      LinearFlowRunnerEngine,
                                      TaskRunnerEngine)
from omnipy.util.helpers import resolve


class PrefectEngine(TaskRunnerEngine,
                    LinearFlowRunnerEngine,
                    DagFlowRunnerEngine,
                    FuncFlowRunnerEngine):
    def _init_engine(self) -> None:
        ...

    def _update_from_config(self) -> None:
        ...

    @classmethod
    def get_config_cls(cls) -> Type[IsPrefectEngineConfig]:
        return PrefectEngineConfig

    # TaskRunnerEngine

    def _init_task(self, task: IsTask, call_func: Callable) -> PrefectTask:

        assert isinstance(self._config, PrefectEngineConfig)
        task_kwargs = dict(
            name=task.name,
            cache_key_fn=task_input_hash if self._config.use_cached_results else None,
            cache_expiration=timedelta(days=1))

        if task.has_coroutine_func():

            @prefect_task(**task_kwargs)
            async def run_task(*inner_args, **inner_kwargs):
                return await call_func(*inner_args, **inner_kwargs)
        else:

            @prefect_task(**task_kwargs)
            def run_task(*inner_args, **inner_kwargs):
                return call_func(*inner_args, **inner_kwargs)

        return run_task

    def _run_task(self, state: PrefectTask, task: IsTask, call_func: Callable, *args,
                  **kwargs) -> Any:

        _prefect_task = state

        if task.in_flow_context:
            return _prefect_task(*args, **kwargs)
        else:
            flow_kwargs = dict(name=task.name)

            if asyncio.iscoroutinefunction(_prefect_task):

                @prefect_flow(**flow_kwargs)
                async def task_flow(*inner_args, **inner_kwargs):
                    return await resolve(_prefect_task(*inner_args, **inner_kwargs))
            else:

                @prefect_flow(**flow_kwargs)
                def task_flow(*inner_args, **inner_kwargs):
                    return _prefect_task(*inner_args, **inner_kwargs)

            return task_flow(*args, **kwargs)

    # LinearFlowRunnerEngine
    def _init_linear_flow(self, linear_flow: IsLinearFlow) -> Any:
        assert isinstance(self._config, PrefectEngineConfig)
        flow_kwargs = dict(name=linear_flow.name,)
        call_func = self.default_linear_flow_run_decorator(linear_flow)

        if linear_flow.has_coroutine_func():

            @prefect_flow(**flow_kwargs)
            async def run_linear_flow(*inner_args, **inner_kwargs):
                with linear_flow.flow_context:
                    return await resolve(call_func(*inner_args, **inner_kwargs))
        else:

            @prefect_flow(**flow_kwargs)
            def run_linear_flow(*inner_args, **inner_kwargs):
                with linear_flow.flow_context:
                    return call_func(*inner_args, **inner_kwargs)

        return run_linear_flow

    def _run_linear_flow(self, state: Any, linear_flow: IsLinearFlow, *args, **kwargs) -> Any:

        _prefect_flow = state

        return _prefect_flow(*args, **kwargs)

    # DagFlowRunnerEngine

    def _init_dag_flow(self, dag_flow: IsDagFlow) -> Any:

        assert isinstance(self._config, PrefectEngineConfig)
        flow_kwargs = dict(name=dag_flow.name,)
        call_func = self.default_dag_flow_run_decorator(dag_flow)

        if dag_flow.has_coroutine_func():

            @prefect_flow(**flow_kwargs)
            async def run_dag_flow(*inner_args, **inner_kwargs):
                with dag_flow.flow_context:
                    return await resolve(call_func(*inner_args, **inner_kwargs))
        else:

            @prefect_flow(**flow_kwargs)
            def run_dag_flow(*inner_args, **inner_kwargs):
                with dag_flow.flow_context:
                    return call_func(*inner_args, **inner_kwargs)

        return run_dag_flow

    def _run_dag_flow(self, state: PrefectFlow, flow: IsDagFlow, *args, **kwargs) -> Any:

        _prefect_flow = state

        return _prefect_flow(*args, **kwargs)

    # FuncFlowRunnerEngine

    def _init_func_flow(self, func_flow: IsFuncFlow, call_func: Callable) -> object:

        assert isinstance(self._config, PrefectEngineConfig)
        flow_kwargs = dict(name=func_flow.name,)

        if func_flow.has_coroutine_func():

            @prefect_flow(**flow_kwargs)
            async def run_func_flow(*inner_args, **inner_kwargs):
                with func_flow.flow_context:
                    return await resolve(call_func(*inner_args, **inner_kwargs))
        else:

            @prefect_flow(**flow_kwargs)
            def run_func_flow(*inner_args, **inner_kwargs):
                with func_flow.flow_context:
                    return call_func(*inner_args, **inner_kwargs)

        return run_func_flow

    def _run_func_flow(self,
                       state: Any,
                       func_flow: IsFuncFlow,
                       call_func: Callable,
                       *args,
                       **kwargs) -> Any:

        _prefect_flow = state

        return _prefect_flow(*args, **kwargs)


# TODO: Refactor redundant code



from dataclasses import dataclass, field
import logging
from sys import stdout
from typing import Any, Optional

from omnipy.api.enums import EngineChoice
from omnipy.api.protocols import (IsEngine,
                                  IsEngineConfig,
                                  IsJobConfig,
                                  IsJobConfigHolder,
                                  IsLocalRunnerConfig,
                                  IsPrefectEngineConfig,
                                  IsRunStateRegistry,
                                  IsRunStateRegistryConfig,
                                  IsRuntime,
                                  IsRuntimeConfig,
                                  IsRuntimeObjects)
from omnipy.compute.job import JobBase
from omnipy.config.engine import LocalRunnerConfig, PrefectEngineConfig
from omnipy.config.job import JobConfig
from omnipy.config.registry import RunStateRegistryConfig
from omnipy.data.serializer import SerializerRegistry
from omnipy.engine.local import LocalRunner
from omnipy.engine.prefect import PrefectEngine
from omnipy.hub.publisher import ConfigPublisher
from omnipy.log.registry import RunStateRegistry
from omnipy.modules.json.serializers import JsonDatasetToTarFileSerializer
from omnipy.modules.pandas.serializers import PandasDatasetToTarFileSerializer
from omnipy.modules.raw.serializers import RawDatasetToTarFileSerializer


def get_default_logger():
    logger = logging.getLogger('omnipy')
    logger.setLevel(logging.INFO)
    for handler in logger.handlers:
        logger.removeHandler(handler)
    logger.addHandler(logging.StreamHandler(stdout))
    return logger


@dataclass
class RuntimeEntry(ConfigPublisher):
    _back: Optional[IsRuntime] = field(default=None, init=False, repr=False)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

        if hasattr(self, key) and not key.startswith('_') and self._back is not None:
            self._back.reset_subscriptions()


def _job_creator_factory():
    return JobBase.job_creator


@dataclass
class RuntimeObjects(RuntimeEntry, ConfigPublisher):
    logger: logging.Logger = field(default_factory=get_default_logger)
    registry: IsRunStateRegistry = field(default_factory=RunStateRegistry)
    job_creator: IsJobConfigHolder = field(default_factory=_job_creator_factory)
    local: IsEngine = field(default_factory=LocalRunner)
    prefect: IsEngine = field(default_factory=PrefectEngine)


@dataclass
class RuntimeConfig(RuntimeEntry, ConfigPublisher):
    job: IsJobConfig = field(default_factory=JobConfig)
    engine: EngineChoice = EngineChoice.LOCAL
    local: IsLocalRunnerConfig = field(default_factory=LocalRunnerConfig)
    prefect: IsPrefectEngineConfig = field(default_factory=PrefectEngineConfig)
    registry: IsRunStateRegistryConfig = field(default_factory=RunStateRegistryConfig)


@dataclass
class Runtime(ConfigPublisher):
    objects: IsRuntimeObjects = field(default_factory=RuntimeObjects)
    config: IsRuntimeConfig = field(default_factory=RuntimeConfig)

    def __post_init__(self):
        super().__init__()

        self.objects._back = self
        self.config._back = self

        self.reset_subscriptions()

    def reset_subscriptions(self):
        self.objects.unsubscribe_all()
        self.config.unsubscribe_all()

        self.objects.subscribe('registry', self.objects.local.set_registry)
        self.objects.subscribe('registry', self.objects.prefect.set_registry)
        self.objects.subscribe('logger', self.objects.registry.set_logger)

        self.objects.subscribe('local', self._update_local_runner_config)
        self.objects.subscribe('prefect', self._update_prefect_engine_config)

        self.config.subscribe('job', self.objects.job_creator.set_config)
        self.config.subscribe('local', self.objects.local.set_config)
        self.config.subscribe('prefect', self.objects.prefect.set_config)
        self.config.subscribe('registry', self.objects.registry.set_config)

        self.config.subscribe('local', self._update_job_creator_engine)
        self.config.subscribe('prefect', self._update_job_creator_engine)
        self.config.subscribe('engine', self._update_job_creator_engine)

    def _get_engine(self, engine_choice: EngineChoice):
        return getattr(self.objects, engine_choice)

    def _get_engine_config(self, engine_choice: EngineChoice):
        return getattr(self.config, engine_choice)

    def _set_engine_config(self, engine_choice: EngineChoice, engine_config: IsEngineConfig):
        return setattr(self.config, engine_choice, engine_config)

    def _new_engine_config_if_new_cls(self, engine: IsEngine, engine_choice: EngineChoice) -> None:
        # TODO: when parsing config from file is implemented, make sure that the new engine
        #       config classes here reparse the config files
        engine_config_cls = engine.get_config_cls()
        if self._get_engine_config(engine_choice).__class__ is not engine_config_cls:
            self._set_engine_config(engine_choice, engine_config_cls())

    def _update_local_runner_config(self, local_runner: IsEngine):
        self._new_engine_config_if_new_cls(local_runner, EngineChoice.LOCAL)

    def _update_prefect_engine_config(self, prefect_engine: IsEngine):
        self._new_engine_config_if_new_cls(prefect_engine, EngineChoice.PREFECT)

    def _update_job_creator_engine(self, _item_changed: Any):
        self.objects.job_creator.set_engine(self._get_engine(self.config.engine))

    def _create_serializer_registry(self):
        registry = SerializerRegistry()
        #
        registry.register(PandasDatasetToTarFileSerializer)
        registry.register(RawDatasetToTarFileSerializer)
        registry.register(JsonDatasetToTarFileSerializer)

        return registry



from abc import ABC, abstractmethod
import inspect
import sys
from types import AsyncGeneratorType, GeneratorType
from typing import Any, Awaitable, Callable, cast

from omnipy.api.enums import RunState
from omnipy.api.protocols import IsDagFlow, IsFuncFlow, IsJob, IsLinearFlow, IsTask
from omnipy.engine.base import Engine


class JobRunnerEngine(Engine, ABC):
    def _register_job_state(self, job: IsJob, state: RunState) -> None:
        if self._registry:
            self._registry.set_job_state(job, state)

    def _decorate_result_with_job_finalization_detector(self, job: IsJob, job_result: object):
        if isinstance(job_result, GeneratorType):
            job_result = cast(GeneratorType, job_result)

            def detect_finished_generator_decorator():
                try:
                    value = yield next(job_result)
                    while True:
                        value = yield job_result.send(value)
                except StopIteration:
                    self._register_job_state(job, RunState.FINISHED)

            return detect_finished_generator_decorator()
        elif isinstance(job_result, AsyncGeneratorType):
            job_result = cast(AsyncGeneratorType, job_result)

            async def detect_finished_async_generator_decorator():
                try:
                    if sys.version_info >= (3, 10):
                        value = yield await anext(job_result)
                    else:
                        value = yield await job_result.__anext__()
                    while True:
                        print(value)
                        value = yield await job_result.asend(value)
                except StopAsyncIteration:
                    self._register_job_state(job, RunState.FINISHED)

            return detect_finished_async_generator_decorator()

        elif inspect.isawaitable(job_result):
            job_result = cast(Awaitable, job_result)

            async def detect_finished_coroutine():
                result = await job_result
                self._register_job_state(job, RunState.FINISHED)
                return result

            return detect_finished_coroutine()
        else:
            self._register_job_state(job, RunState.FINISHED)
            return job_result


class TaskRunnerEngine(JobRunnerEngine):
    def apply_task_decorator(self, task: IsTask, job_callback_accept_decorator: Callable) -> None:
        def _task_decorator(call_func: Callable) -> Callable:
            self._register_job_state(task, RunState.INITIALIZED)
            state = self._init_task(task, call_func)

            def _task_runner_call_func(*args: Any, **kwargs: Any) -> Any:
                self._register_job_state(task, RunState.RUNNING)
                task_result = self._run_task(state, task, call_func, *args, **kwargs)
                return self._decorate_result_with_job_finalization_detector(task, task_result)

            return _task_runner_call_func

        job_callback_accept_decorator(_task_decorator)

    @abstractmethod
    def _init_task(self, task: IsTask, call_func: Callable) -> Any:
        ...

    @abstractmethod
    def _run_task(self, state: Any, task: IsTask, call_func: Callable, *args, **kwargs) -> Any:
        ...


class LinearFlowRunnerEngine(JobRunnerEngine):
    def apply_linear_flow_decorator(self,
                                    linear_flow: IsLinearFlow,
                                    job_callback_accept_decorator: Callable) -> None:
        def _linear_flow_decorator(call_func: Callable) -> Callable:
            self._register_job_state(linear_flow, RunState.INITIALIZED)
            state = self._init_linear_flow(linear_flow)

            def _linear_flow_runner_call_func(*args: object, **kwargs: object) -> Any:
                self._register_job_state(linear_flow, RunState.RUNNING)
                flow_result = self._run_linear_flow(state, linear_flow, *args, **kwargs)
                return self._decorate_result_with_job_finalization_detector(
                    linear_flow, flow_result)

            return _linear_flow_runner_call_func

        job_callback_accept_decorator(_linear_flow_decorator)

    @staticmethod
    def default_linear_flow_run_decorator(linear_flow: IsLinearFlow) -> Any:
        def _inner_run_linear_flow(*args: object, **kwargs: object):

            result = None
            with linear_flow.flow_context:
                for i, job in enumerate(linear_flow.task_templates):
                    # TODO: Better handling of kwargs
                    if i == 0:
                        result = job(*args, **kwargs)
                    else:
                        result = job(*args)

                    args = (result,)
            return result

        return _inner_run_linear_flow

    @abstractmethod
    def _init_linear_flow(self, linear_flow: IsLinearFlow) -> Any:
        ...

    @abstractmethod
    def _run_linear_flow(self, state: Any, linear_flow: IsLinearFlow, *args, **kwargs) -> Any:
        ...


class DagFlowRunnerEngine(JobRunnerEngine):
    def apply_dag_flow_decorator(self, dag_flow: IsDagFlow,
                                 job_callback_accept_decorator: Callable) -> None:
        def _dag_flow_decorator(call_func: Callable) -> Callable:
            self._register_job_state(dag_flow, RunState.INITIALIZED)
            state = self._init_dag_flow(dag_flow)

            def _dag_flow_runner_call_func(*args: object, **kwargs: object) -> Any:
                self._register_job_state(dag_flow, RunState.RUNNING)
                flow_result = self._run_dag_flow(state, dag_flow, *args, **kwargs)
                return self._decorate_result_with_job_finalization_detector(dag_flow, flow_result)

            return _dag_flow_runner_call_func

        job_callback_accept_decorator(_dag_flow_decorator)

    @staticmethod
    def default_dag_flow_run_decorator(dag_flow: IsDagFlow) -> Any:
        def _inner_run_dag_flow(*args: object, **kwargs: object):
            results = {}
            result = None
            with dag_flow.flow_context:
                for i, job in enumerate(dag_flow.task_templates):
                    if i == 0:
                        results = dag_flow.get_call_args(*args, **kwargs)

                    param_keys = set(inspect.signature(job).parameters.keys())

                    # TODO: Refactor to remove dependency
                    #       Also, add test for not allowing override of fixed_params
                    if hasattr(job, 'param_key_map'):
                        for key, val in job.param_key_map.items():
                            if key in param_keys:
                                param_keys.remove(key)
                                param_keys.add(val)

                    if hasattr(job, 'fixed_params'):
                        for key in job.fixed_params.keys():
                            if key in param_keys:
                                param_keys.remove(key)

                    params = {key: val for key, val in results.items() if key in param_keys}
                    result = job(**params)

                    if isinstance(result, dict) and len(result) > 0:
                        results.update(result)
                    else:
                        results[job.name] = result
            return result

        return _inner_run_dag_flow

    @abstractmethod
    def _init_dag_flow(self, dag_flow: IsDagFlow) -> Any:
        ...

    @abstractmethod
    def _run_dag_flow(self, state: Any, dag_flow: IsDagFlow, *args, **kwargs) -> Any:
        ...


class FuncFlowRunnerEngine(JobRunnerEngine):
    def apply_func_flow_decorator(self,
                                  func_flow: IsFuncFlow,
                                  job_callback_accept_decorator: Callable) -> None:
        def _func_flow_decorator(call_func: Callable) -> Callable:
            self._register_job_state(func_flow, RunState.INITIALIZED)
            state = self._init_func_flow(func_flow, call_func)

            def _func_flow_runner_call_func(*args: object, **kwargs: object) -> Any:
                self._register_job_state(func_flow, RunState.RUNNING)
                with func_flow.flow_context:
                    flow_result = self._run_func_flow(state, func_flow, call_func, *args, **kwargs)
                    return self._decorate_result_with_job_finalization_detector(
                        func_flow, flow_result)

            return _func_flow_runner_call_func

        job_callback_accept_decorator(_func_flow_decorator)

    @abstractmethod
    def _init_func_flow(self, func_flow: IsFuncFlow, call_func: Callable) -> object:
        ...

    @abstractmethod
    def _run_func_flow(self,
                       state: Any,
                       func_flow: IsFuncFlow,
                       call_func: Callable,
                       *args,
                       **kwargs) -> Any:
        ...



from __future__ import annotations

from typing import cast, Type

from omnipy.api.protocols import IsEngine, IsFuncJobTemplateCallable, IsTaskRunnerEngine
from omnipy.compute.func_job import FuncArgJobBase
from omnipy.compute.job import Job, JobTemplate
from omnipy.util.callable_decorator_cls import callable_decorator_cls

# class TaskInit(FuncArgJobBase):
#     def __init__(
#         self,
#         job_func: Callable,
#         name: Optional[str] = None,
#         iterate_over_data_files: bool = False,
#         fixed_params: Optional[Mapping[str, object]] = None,
#         param_key_map: Optional[Mapping[str, str]] = None,
#         result_key: Optional[str] = None,
#         persist_outputs: Optional[PersistOutputsOptions] = None,
#         restore_outputs: Optional[RestoreOutputsOptions] = None,
#         **kwargs: object,
#     ):
#         ...
#
#     #     super().__init__(
#     #         job_func,
#     #         **remove_none_vals(
#     #             name=name,
#     #             iterate_over_data_files=iterate_over_data_files,
#     #             fixed_params=fixed_params,
#     #             param_key_map=param_key_map,
#     #             result_key=result_key,
#     #             persist_outputs=persist_outputs,
#     #             restore_outputs=restore_outputs,
#     #             **kwargs,
#     #         ))


def task_template_callable_decorator_cls(
        cls: Type['TaskTemplate']) -> IsFuncJobTemplateCallable['TaskTemplate']:
    return cast(IsFuncJobTemplateCallable['TaskTemplate'], callable_decorator_cls(cls))


class TaskBase:
    ...


@task_template_callable_decorator_cls
class TaskTemplate(JobTemplate, TaskBase, FuncArgJobBase):
    @classmethod
    def _get_job_subcls_for_apply(cls) -> Type[Task]:
        return Task


class Task(Job, TaskBase, FuncArgJobBase):
    def _apply_engine_decorator(self, engine: IsEngine) -> None:
        # self._check_engine(IsTaskRunnerEngine)
        if self.engine:
            engine = cast(IsTaskRunnerEngine, self.engine)
            engine.apply_task_decorator(self, self._accept_call_func_decorator)

    @classmethod
    def _get_job_template_subcls_for_revise(cls) -> Type[TaskTemplate]:
        return TaskTemplate


# TODO: Would we need the possibility to refine task templates by adding new task parameters?



import os

from omnipy.compute.task import TaskTemplate
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

from .models import JsonDataset


@TaskTemplate()
def import_directory(directory: str) -> JsonDataset:
    raw_dataset = Dataset[Model[str]]()
    for import_filename in os.listdir(directory):
        print(import_filename)
        if import_filename.endswith('.json'):
            with open(os.path.join(directory, import_filename)) as open_file:
                dataset_name = import_filename.split('.')[0]
                print(dataset_name)
                raw_dataset[dataset_name] = open_file.read()

    dataset = JsonDataset()
    dataset.from_json(raw_dataset.to_data())
    return dataset



from typing import Protocol, Type, TypeVar, Union

from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

InputT = TypeVar('InputT', bound=object)
ModelInputT = TypeVar('ModelInputT', bound=Model)
ReturnT = TypeVar('ReturnT', bound=object)
ModelReturnT = TypeVar('ModelReturnT', bound=Model)
InputTypeT = Union[InputT, ModelInputT]
InputDatasetT = Union[Dataset[Model[InputT]], Dataset[ModelInputT]]
ReturnDatasetT = Union[Dataset[Model[ReturnT]], Dataset[ModelReturnT]]


class IsIterateInnerCallable(Protocol):
    def __call__(
        self,
        data_file: Union[InputT, ModelInputT],
        *args: object,
        **kwargs: object,
    ) -> Union[ReturnT, ModelReturnT]:
        ...

    __name__: str



from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Optional

from omnipy.api.enums import ConfigPersistOutputsOptions, ConfigRestoreOutputsOptions


def _get_persist_data_dir_path():
    return str(Path.cwd().joinpath(Path('data')))


@dataclass
class JobConfig:
    persist_outputs: ConfigPersistOutputsOptions = \
        ConfigPersistOutputsOptions.ENABLE_FLOW_AND_TASK_OUTPUTS
    restore_outputs: ConfigRestoreOutputsOptions = \
        ConfigRestoreOutputsOptions.DISABLED
    persist_data_dir_path: Path = field(default_factory=_get_persist_data_dir_path)
    datetime_of_nested_context: Optional[datetime] = None



import asyncio
from typing import Callable, Dict, Tuple

from omnipy.api.protocols import GeneralDecorator
from omnipy.compute.job import JobBase
# from omnipy.compute.job_types import FuncJobTemplateT, JobBaseT, JobT, JobTemplateT
from omnipy.compute.mixins.func_signature import SignatureFuncJobBaseMixin
from omnipy.compute.mixins.iterate import IterateFuncJobBaseMixin
from omnipy.compute.mixins.params import ParamsFuncJobBaseMixin
from omnipy.compute.mixins.result_key import ResultKeyFuncJobBaseMixin
from omnipy.compute.mixins.serialize import SerializerFuncJobBaseMixin


class PlainFuncArgJobBase(JobBase):
    def __init__(self, job_func: Callable, *args: object, **kwargs: object) -> None:
        # JobBase.__init__(self, job_func, *args, **kwargs)
        self._job_func = job_func

    def _get_init_args(self) -> Tuple[object, ...]:
        return self._job_func,

    def has_coroutine_func(self) -> bool:
        return asyncio.iscoroutinefunction(self._job_func)

    def _call_job(self, *args: object, **kwargs: object) -> object:
        """To be overloaded by mixins"""
        return self._call_func(*args, **kwargs)

    def _call_func(self, *args: object, **kwargs: object) -> object:
        """
        To be decorated by job runners and mixins that need early application. Should not
        be overloaded using inheritance. The method _accept_call_func_decorator accepts
        decorators.
        """
        return self._job_func(*args, **kwargs)

    def _accept_call_func_decorator(self, call_func_decorator: GeneralDecorator) -> None:
        self._call_func = call_func_decorator(self._call_func)


# Extra level needed for mixins to be able to overload _call_job (and possibly other methods)
class FuncArgJobBase(PlainFuncArgJobBase):
    ...


FuncArgJobBase.accept_mixin(SignatureFuncJobBaseMixin)
FuncArgJobBase.accept_mixin(IterateFuncJobBaseMixin)
FuncArgJobBase.accept_mixin(SerializerFuncJobBaseMixin)
FuncArgJobBase.accept_mixin(ResultKeyFuncJobBaseMixin)
FuncArgJobBase.accept_mixin(ParamsFuncJobBaseMixin)

# class FuncJobTemplate(FuncArgJobBase, JobTemplate, ABC):
#     def refine(self,
#                update: bool = True,
#                name: Optional[str] = None,
#                iterate_over_data_files: bool = False,
#                fixed_params: Optional[Mapping[str, object]] = None,
#                param_key_map: Optional[Mapping[str, str]] = None,
#                result_key: Optional[str] = None,
#                persist_outputs: Optional[PersistOutputsOptions] = None,
#                restore_outputs: Optional[RestoreOutputsOptions] = None,
#                **kwargs: object):  # -> FuncJobTemplateT:
#
#         return self._refine(
#             self,
#             update=update,
#             **remove_none_vals(
#                 name=name,
#                 iterate_over_data_files=iterate_over_data_files,
#                 fixed_params=fixed_params,
#                 param_key_map=param_key_map,
#                 result_key=result_key,
#                 persist_outputs=persist_outputs,
#                 restore_outputs=restore_outputs,
#                 **kwargs,
#             ))



from datetime import datetime
import os
from pathlib import Path
import tarfile
from typing import Optional

from omnipy.api.enums import ConfigPersistOutputsOptions as ConfigPersistOpts
from omnipy.api.enums import ConfigRestoreOutputsOptions as ConfigRestoreOpts
from omnipy.api.enums import PersistOutputsOptions, RestoreOutputsOptions
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model
from omnipy.data.serializer import SerializerRegistry
from omnipy.modules.json.serializers import JsonDatasetToTarFileSerializer
from omnipy.modules.pandas.serializers import PandasDatasetToTarFileSerializer
from omnipy.modules.raw.serializers import RawDatasetToTarFileSerializer

PersistOpts = PersistOutputsOptions
RestoreOpts = RestoreOutputsOptions


class SerializerFuncJobBaseMixin:
    def __init__(self,
                 *,
                 persist_outputs: Optional[PersistOutputsOptions] = None,
                 restore_outputs: Optional[RestoreOutputsOptions] = None):

        # TODO: Possibly reimplement logic using a state machine, e.g. "transitions" package
        if persist_outputs is None:
            self._persist_outputs = PersistOpts.FOLLOW_CONFIG if self._has_job_config else None
        else:
            self._persist_outputs = PersistOpts(persist_outputs)

        if restore_outputs is None:
            self._restore_outputs = RestoreOpts.FOLLOW_CONFIG if self._has_job_config else None
        else:
            self._restore_outputs = RestoreOpts(restore_outputs)

        self._serializer_registry = self._create_serializer_registry()

    def _create_serializer_registry(self):
        # TODO: store in runtime, to remove dependencies
        registry = SerializerRegistry()

        registry.register(PandasDatasetToTarFileSerializer)
        registry.register(RawDatasetToTarFileSerializer)
        registry.register(JsonDatasetToTarFileSerializer)

        return registry

    @property
    def _has_job_config(self) -> bool:
        return self.config is not None

    @property
    def persist_outputs(self) -> Optional[PersistOutputsOptions]:
        return self._persist_outputs

    @property
    def restore_outputs(self) -> Optional[RestoreOutputsOptions]:
        return self._restore_outputs

    @property
    def will_persist_outputs(self) -> PersistOutputsOptions:
        if not self._has_job_config or self._persist_outputs is not PersistOpts.FOLLOW_CONFIG:
            return self._persist_outputs if self._persist_outputs is not None \
                    else PersistOpts.DISABLED
        else:
            # TODO: Refactor using Flow and Task Mixins
            from omnipy.compute.flow import FlowBase
            from omnipy.compute.task import TaskBase

            config_persist_opt = self.config.persist_outputs

            if config_persist_opt == ConfigPersistOpts.ENABLE_FLOW_OUTPUTS:
                return PersistOpts.ENABLED if isinstance(self, FlowBase) else PersistOpts.DISABLED
            elif config_persist_opt == ConfigPersistOpts.ENABLE_FLOW_AND_TASK_OUTPUTS:
                return PersistOpts.ENABLED \
                        if any(isinstance(self, cls) for cls in (FlowBase, TaskBase)) \
                        else PersistOpts.DISABLED
            else:
                assert config_persist_opt == ConfigPersistOpts.DISABLED
                return PersistOpts.DISABLED

    @property
    def will_restore_outputs(self) -> RestoreOutputsOptions:
        if not self._has_job_config or self._restore_outputs is not RestoreOpts.FOLLOW_CONFIG:
            return self._restore_outputs if self._restore_outputs is not None \
                    else RestoreOpts.DISABLED
        else:
            config_restore_opt = self.config.restore_outputs

            if config_restore_opt == ConfigRestoreOpts.AUTO_ENABLE_IGNORE_PARAMS:
                return RestoreOpts.AUTO_ENABLE_IGNORE_PARAMS
            assert config_restore_opt == ConfigRestoreOpts.DISABLED
            return RestoreOpts.DISABLED

    def _call_job(self, *args: object, **kwargs: object) -> object:
        if self.will_restore_outputs in [
                RestoreOpts.AUTO_ENABLE_IGNORE_PARAMS, RestoreOpts.FORCE_ENABLE_IGNORE_PARAMS
        ]:
            try:
                return self._deserialize_and_restore_outputs()
            except Exception:
                if self.will_restore_outputs is RestoreOpts.FORCE_ENABLE_IGNORE_PARAMS:
                    raise

        results = super()._call_job(*args, **kwargs)

        if self.will_persist_outputs is PersistOpts.ENABLED:
            if isinstance(results, Dataset):
                self._serialize_and_persist_outputs(results)
            else:
                self._log_message(f'Results of {self.unique_name} is not a Dataset and cannot '
                                  f'be automatically serialized and persisted!')

        return results

    def _serialize_and_persist_outputs(self, results: Dataset):
        datetime_str = self._generate_datetime_str()
        output_path = Path(self.config.persist_data_dir_path).joinpath(datetime_str)

        if not os.path.exists(output_path):
            os.makedirs(output_path)

        num_cur_files = len(os.listdir(output_path))
        job_name = self._create_job_name()

        file_path = output_path.joinpath(f'{num_cur_files:02}_{job_name}.tar.gz')

        parsed_dataset, serializer = \
            self._serializer_registry.auto_detect_tar_file_serializer(results)

        if serializer is None:
            self._log_message(f'Unable to find a serializer for results of job "{self.name}", '
                              f'with data type "{type(results)}". Will abort persisting results...')
        else:
            self._log_message(
                f'Writing dataset as a gzipped tarpack to "{os.path.abspath(file_path)}"')

            with open(file_path, 'wb') as tarfile:
                tarfile.write(serializer.serialize(parsed_dataset))

    def _create_job_name(self):
        return '_'.join(self.unique_name.split('-')[:-2])

    def _generate_datetime_str(self):
        if self.time_of_cur_toplevel_flow_run:
            run_time = self.time_of_cur_toplevel_flow_run
        else:
            if hasattr(self, 'time_of_last_run'):
                run_time = self.time_of_last_run
            else:
                run_time = datetime.now()
        datetime_str = run_time.strftime('%Y_%m_%d-%H_%M_%S')
        return datetime_str

    # TODO: Refactor
    def _deserialize_and_restore_outputs(self) -> Dataset:
        output_path = Path(self.config.persist_data_dir_path)
        if os.path.exists(output_path):
            sorted_date_dirs = list(sorted(os.listdir(output_path)))
            if len(sorted_date_dirs) > 0:
                last_dir = sorted_date_dirs[-1]
                last_dir_path = output_path.joinpath(last_dir)
                for job_output_name in reversed(sorted(os.listdir(last_dir_path))):
                    name_part_of_filename = job_output_name[3:-7]
                    if name_part_of_filename == self._create_job_name():
                        tar_file_path = last_dir_path.joinpath(job_output_name)
                        with tarfile.open(tar_file_path, 'r:gz') as tarfile_obj:
                            file_suffixes = set(fn.split('.')[-1] for fn in tarfile_obj.getnames())
                        if len(file_suffixes) != 1:
                            self._log_message(f'Tar archive contains files with different or '
                                              f'no file suffixes: {file_suffixes}. Serializer '
                                              f'cannot be uniquely determined. Aborting '
                                              f'restore.')
                        else:
                            file_suffix = file_suffixes.pop()
                            serializers = self._serializer_registry.\
                                detect_tar_file_serializers_from_file_suffix(file_suffix)
                            if len(serializers) == 0:
                                self._log_message(
                                    f'No serializer for file suffix "{file_suffix}" can be'
                                    f'determined. Aborting restore.')
                            else:
                                self._log_message(f'Reading dataset from a gzipped tarpack at'
                                                  f' "{os.path.abspath(tar_file_path)}"')

                                serializer = serializers[0]
                                with open(tar_file_path, 'rb') as tarfile_binary:
                                    dataset = serializer.deserialize(tarfile_binary.read())
                                if dataset.get_model_class() is \
                                        self.return_type().get_model_class():
                                    return dataset
                                else:
                                    try:
                                        new_dataset = self.return_type()
                                        if new_dataset.get_model_class() is Model[str]:
                                            new_dataset.from_data(dataset.to_json())
                                        else:
                                            new_dataset.from_json(dataset.to_data())
                                        return new_dataset
                                    except:
                                        return dataset

        raise RuntimeError('No persisted output')

    @classmethod
    def _log_message(cls, log_message: str) -> None:
        from omnipy import runtime
        if runtime:
            runtime.objects.registry.log(datetime.now(), log_message)

    # TODO: Refactor logging as a general mixin and add it to at least RunStateRegistry (as now)
    #       and also JobCreator, possibly engines, and more. Use RuntimeConfig to subscribe to
    #       loggers, as now.



import inspect
import locale as pkg_locale
from typing import Dict, get_args, get_origin, Iterable, Mapping, Tuple, Type, Union

from typing_inspect import get_generic_bases, is_generic_type

DictT = Union[Mapping[object, object], Iterable[Tuple[object, object]]]


def create_merged_dict(dict_1: DictT, dict_2: DictT) -> Dict[object, object]:
    merged_dict = dict(dict_1)
    dict_2_cast = dict(dict_2)
    merged_dict |= dict_2_cast
    return merged_dict


def remove_none_vals(**kwargs: object) -> Dict[object, object]:
    return {key: val for key, val in kwargs.items() if val is not None}


def get_datetime_format(locale: Union[str, Tuple[str, str]] = '') -> str:
    pkg_locale.setlocale(pkg_locale.LC_ALL, locale)

    if hasattr(pkg_locale, 'nl_langinfo'):  # noqa
        datetime_format = pkg_locale.nl_langinfo(pkg_locale.D_T_FMT)
    else:
        datetime_format = '%a %b %e %X %Y'
    return datetime_format


async def resolve(val):
    return await val if inspect.isawaitable(val) else val


def repr_max_len(data: object, max_len: int = 200):
    repr_str = repr(data)
    return f'{repr_str[:max_len]}...' if len(repr_str) > max_len else repr_str


def get_bases(cls):
    return get_generic_bases(cls) if is_generic_type(cls) else cls.__bases__


def generic_aware_issubclass(cls, cls_or_tuple):
    try:
        return issubclass(cls, cls_or_tuple)
    except TypeError:
        return issubclass(get_origin(cls), cls_or_tuple)


def transfer_generic_args_to_cls(to_cls, from_generic_type):
    try:
        return to_cls[get_args(from_generic_type)]
    except (TypeError, AttributeError):
        return to_cls



from typing import Dict, List, Union

from omnipy.data.model import Model

from ..json.models import JsonDict, JsonList


class TableOfStrings(Model[List[Dict[str, str]]]):
    ...


class JsonTableOfStrings(Model[JsonList[JsonDict[str]]]):
    ...


class TableOfStringsAndLists(Model[List[Dict[str, Union[str, List[str]]]]]):
    ...



from abc import ABC, abstractmethod
from collections import UserDict
from typing import Any

from pydantic import BaseModel, validate_model


class OldDataset(ABC, UserDict, BaseModel):
    def __init__(self):
        BaseModel.__init__(self, data={})
        UserDict.__init__(self, {})

    @abstractmethod
    def __setitem__(self, obj_type: str, data_obj: Any) -> None:
        pass


def validate(model: BaseModel):
    *_, validation_error = validate_model(model.__class__, model.__dict__)
    if validation_error:
        raise validation_error



from inspect import Parameter, signature
from typing import Callable, cast, Type

from omnipy.compute.mixins.mixin_types import (InputDatasetT,
                                               InputTypeT,
                                               IsIterateInnerCallable,
                                               ReturnDatasetT)
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

# Functions


def _check_job_func_parameters(job_func: Callable) -> None:
    params = list(signature(job_func).parameters.values())

    if len(params) == 0 or \
            params[0].default != Parameter.empty or \
            params[0].kind not in [Parameter.POSITIONAL_ONLY,
                                   Parameter.POSITIONAL_OR_KEYWORD]:
        raise AttributeError('Parameter "iterate_over_data_files" is set to True, '
                             'but the job function has no arguments without default values. '
                             'Such a first argument will be replaced with a corresponding '
                             'Dataset arg to be iterated over')


def _create_dataset_cls(data_file_type: InputTypeT) -> Type[InputDatasetT]:
    return Dataset[data_file_type] if issubclass(data_file_type, Model) \
        else Dataset[Model[data_file_type]]


def _generate_new_signature(job_func: Callable):
    func_signature = signature(job_func)
    params = list(func_signature.parameters.values())
    data_param = params[0]
    rest_params = params[1:]

    dataset_cls = _create_dataset_cls(data_param.annotation)
    dataset_param = data_param.replace(name='dataset', annotation=dataset_cls)
    out_dataset_cls = _create_dataset_cls(func_signature.return_annotation)

    return func_signature.replace(
        parameters=[dataset_param] + rest_params,
        return_annotation=out_dataset_cls,
    )


# Classes


class IterateFuncJobBaseMixin:
    # Requires FuncSignatureJobBaseMixin

    def __init__(self, *, iterate_over_data_files: bool = False):
        self._iterate_over_data_files = iterate_over_data_files

        if not isinstance(self.iterate_over_data_files, bool):
            raise TypeError(
                'Value of "iterate_over_data_files" parameter must be bool (True/False), '
                f'not "{iterate_over_data_files}"')

        if iterate_over_data_files:
            job_func = self._job_func
            if job_func.__name__ != '_omnipy_iterate_func':

                def _iterate_over_data_files_decorator(call_func: Callable):
                    def _omnipy_iterate_func(
                        dataset: InputDatasetT,
                        *args: object,
                        **kwargs: object,
                    ) -> ReturnDatasetT:
                        inner_func: IsIterateInnerCallable = \
                            cast(IsIterateInnerCallable, call_func)

                        return_type = signature(job_func).return_annotation
                        out_dataset_cls = _create_dataset_cls(return_type)
                        out_dataset = out_dataset_cls()

                        for title, data_file in dataset.items():
                            out_dataset[title] = inner_func(data_file, *args, **kwargs)

                        return out_dataset

                    return _omnipy_iterate_func

                _check_job_func_parameters(job_func)
                self._accept_call_func_decorator(_iterate_over_data_files_decorator)

                new_signature = _generate_new_signature(job_func)
                self._update_func_signature(new_signature)

    @property
    def iterate_over_data_files(self) -> bool:
        return self._iterate_over_data_files



from typing import Any, Dict, IO, Type

from omnipy.data.dataset import Dataset
from omnipy.data.model import Model
from omnipy.data.serializer import TarFileSerializer


class RawDatasetToTarFileSerializer(TarFileSerializer):
    @classmethod
    def get_supported_dataset_type(cls) -> Type[Dataset]:
        return Dataset[Model[str]]

    @classmethod
    def get_output_file_suffix(cls) -> str:
        return 'raw'

    @classmethod
    def serialize(cls, dataset: Dataset[Model[str]]) -> bytes:
        def raw_encode_func(contents: str) -> bytes:
            return contents.encode('utf8')

        return cls.create_tarfile_from_dataset(dataset, data_encode_func=raw_encode_func)

    @classmethod
    def deserialize(cls, tarfile_bytes: bytes) -> Dataset[Model[str]]:
        dataset = Dataset[Model[str]]()

        def raw_decode_func(file_stream: IO[bytes]) -> str:
            return file_stream.read().decode('utf8')

        def python_dictify_object(obj_type: str, obj_val: Any) -> Dict:
            return {obj_type: obj_val}

        cls.create_dataset_from_tarfile(
            dataset,
            tarfile_bytes,
            data_decode_func=raw_decode_func,
            dictify_object_func=python_dictify_object,
            import_method='from_data')  # noqa

        return dataset



from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from functools import update_wrapper
from types import MappingProxyType
from typing import Any, Dict, Hashable, Optional, Tuple

from omnipy.api.exceptions import JobStateException
from omnipy.api.protocols import IsEngine, IsJobConfig
from omnipy.compute.job_creator import JobBaseMeta
from omnipy.compute.mixins.name import NameJobBaseMixin, NameJobMixin
from omnipy.util.helpers import create_merged_dict
from omnipy.util.mixin import DynamicMixinAcceptor


class JobBase(DynamicMixinAcceptor, metaclass=JobBaseMeta):
    def __init__(self, *args: object, name: Optional[str] = None, **kwargs: object):
        # super().__init__()

        # TODO: refactor using state machine

        if not isinstance(self, JobTemplate) and not isinstance(self, Job):
            raise JobStateException('JobBase and subclasses not inheriting from JobTemplate '
                                    'or Job are not directly instantiatable')

        from_apply = hasattr(self, '_from_apply') and self._from_apply is True
        if isinstance(self, Job) and not from_apply:
            raise JobStateException('Job should only be instantiated using the "apply()" method of '
                                    'an instance of JobTemplate (or one of its subclasses)')

    @classmethod
    def _create_job_template(cls, *args: object, **kwargs: object):  # -> JobTemplateT:
        if len(args) >= 1:
            if not callable(args[0]):
                raise TypeError(f'First argument of a job is assumed to be a callable, '
                                f'not: {args[0]}')
            job_func = args[0]
            args = args[1:]
            return cls(*args, **kwargs)(job_func)
        else:
            return cls(*args, **kwargs)

    @classmethod
    def _create_job(cls, *args: object, **kwargs: object):  # -> Job[JobBaseT, JobTemplateT]:
        if cls.__new__ is object.__new__:
            job_obj = cls.__new__(cls)
        else:
            job_obj = cls.__new__(cls, *args, **kwargs)

        # TODO: refactor using state machine
        job_obj._from_apply = True
        job_obj.__init__(*args, **kwargs)
        job_obj._from_apply = False

        return job_obj

    def _apply(self):  # -> JobT:
        job_cls = self._get_job_subcls_for_apply()
        job = job_cls.create_job(*self._get_init_args(), **self._get_init_kwargs())
        update_wrapper(job, self, updated=[])
        job._apply_engine_decorator(self.engine)
        return job

    def _refine(self, *args: object, update: bool = True, **kwargs: object):  # -> JobTemplateT:
        refine_kwargs = kwargs.copy()

        if update:
            for key, cur_val in self._get_init_kwargs().items():
                if key in refine_kwargs:
                    try:
                        new_val = create_merged_dict(cur_val, refine_kwargs[key])
                    except (TypeError, ValueError):
                        new_val = refine_kwargs[key]
                else:
                    new_val = cur_val
                refine_kwargs[key] = new_val

        init_args = self._get_init_args()
        refine_args = [init_args[0]] if len(init_args) >= 1 else []
        refine_args += args if args else init_args[1:]

        return self.create_job_template(*refine_args, **refine_kwargs)

    def _revise(self):  # -> JobTemplateT:
        job_template_cls = self._get_job_template_subcls_for_revise()
        job_template = job_template_cls.create_job_template(*self._get_init_args(),
                                                            **self._get_init_kwargs())
        update_wrapper(job_template, self, updated=[])
        return job_template

    def _get_init_args(self) -> Tuple[object, ...]:
        return ()

    def _get_init_kwargs(self) -> Dict[str, Any]:
        kwarg_keys = list(self._mixin_init_kwarg_params_including_bases.keys())
        for key in kwarg_keys:
            attribute = getattr(self.__class__, key)

            if not isinstance(attribute, property):
                raise TypeError(f'{self.__class__.__name__} attribute "{key}" is not a property')

            if attribute.fset is not None:
                raise TypeError(f'{self.__class__.__name__} attribute "{key}" is not a property')

            value = getattr(self, key)
            if not isinstance(value, Hashable) and not isinstance(value, MappingProxyType):
                raise TypeError(f'Value of {self.__class__.__name__} attribute "{key}" is mutable')

        return {key: getattr(self, key) for key in kwarg_keys}

    @classmethod
    # @abstractmethod
    def _get_job_template_subcls_for_revise(cls):  # -> Type[JobTemplateT]:
        return JobTemplate

    @classmethod
    # @abstractmethod
    def _get_job_subcls_for_apply(cls):  # -> Type[JobT]:
        return Job

    def _call_job_template(self, *args: object, **kwargs: object) -> object:
        if self.in_flow_context:
            return self.run(*args, **kwargs)
        raise TypeError(f"'{self.__class__.__name__}' object is not callable")

    def _call_job(self, *args: object, **kwargs: object) -> object:
        pass

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, JobBase):
            return NotImplemented
        return self._get_init_args() == other._get_init_args() \
            and self._get_init_kwargs() == other._get_init_kwargs()

    def _check_engine(self, engine_protocol: IsEngine):
        if self.engine is None or not isinstance(self.engine, engine_protocol):
            raise RuntimeError(f'Engine "{self.engine}" does not support '
                               f'job runner protocol: {engine_protocol.__name__}')

    @property
    def config(self) -> Optional[IsJobConfig]:
        return self.__class__.job_creator.config

    @property
    def engine(self) -> Optional[IsEngine]:
        return self.__class__.engine

    @property
    def in_flow_context(self) -> bool:
        return self.__class__.job_creator.nested_context_level > 0


class JobTemplate:
    def __init__(self, *args, **kwargs):
        if JobBase not in self.__class__.__mro__:
            raise TypeError('JobTemplate is not meant to be instantiated outside the context '
                            'of a JobBase subclass.')

    @classmethod
    def create_job_template(cls, *args: object, **kwargs: object):  # -> JobTemplateT:
        return cls._create_job_template(*args, **kwargs)

    def run(self, *args: object, **kwargs: object) -> object:
        return self.apply()(*args, **kwargs)

    def apply(self):  # -> JobT:
        return self._apply()

    def refine(self, *args: object, update: bool = True, **kwargs: object):  # -> JobTemplateT:
        return self._refine(*args, update=update, **kwargs)

    def __call__(self, *args: object, **kwargs: object) -> object:
        return self._call_job_template(*args, **kwargs)


class Job(DynamicMixinAcceptor):
    def __init__(self, *args, **kwargs):
        if JobBase not in self.__class__.__mro__:
            raise TypeError('Job is not meant to be instantiated outside the context '
                            'of a JobBase subclass.')

    @abstractmethod
    def _apply_engine_decorator(self, engine: IsEngine) -> None:
        ...

    @property
    def time_of_cur_toplevel_flow_run(self) -> datetime:
        return self.__class__.job_creator.time_of_cur_toplevel_nested_context_run

    # def __new__(cls, *args: Any, **kwargs: Any):
    #     super().__new__(cls, *args, **kwargs)
    #     raise RuntimeError('Job should only be instantiated using the "apply()" method of '
    #                        'an instance of JobTemplate (or one of its subclasses)')

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)

    @classmethod
    def create_job(cls, *args: object, **kwargs: object):
        return cls._create_job(*args, **kwargs)

    def revise(self):  # -> JobTemplateT:
        return self._revise()

    def __call__(self, *args: object, **kwargs: object) -> object:
        return self._call_job(*args, **kwargs)


# TODO: Change JobBase and friends into Generics such as one can annotated with
#       e.g. 'TaskTemplate[[int], int]' instead of just 'TaskTemplate'

JobBase.accept_mixin(NameJobBaseMixin)
Job.accept_mixin(NameJobMixin)



from typing import Any, Dict, Iterable, List

from omnipy.data.dataset import Dataset
from omnipy.data.model import Model, ROOT_KEY

from . import pd


class PandasModel(Model[pd.DataFrame]):
    @classmethod
    def _parse_data(cls, data: pd.DataFrame) -> pd.DataFrame:
        cls._data_column_names_are_strings(data)
        cls._data_not_empty_object(data)
        return data

    @staticmethod
    def _data_column_names_are_strings(data: pd.DataFrame) -> None:
        for column in data.columns:
            assert isinstance(column, str)

    @staticmethod
    def _data_not_empty_object(data: pd.DataFrame) -> None:
        assert not any(data.isna().all(axis=1))

    def dict(self, *args, **kwargs) -> Dict[Any, Any]:
        return super().dict(*args, **kwargs)[ROOT_KEY].to_dict(orient='records')  # noqa

    def from_data(self, value: Iterable[Any]) -> None:
        self.contents = self._convert_ints_to_nullable_ints(pd.DataFrame(value))

    def from_json(self, value: str) -> None:
        self.contents = self._convert_ints_to_nullable_ints(pd.read_json(value))

    @classmethod
    def _convert_ints_to_nullable_ints(cls, dataframe: pd.DataFrame) -> pd.DataFrame:
        for key, col in dataframe.items():
            if col.dtype == 'int64' or (col.dtype == 'float64' and col.isna().any()):
                try:
                    dataframe[key] = col.astype(float).astype('Int64')
                except TypeError:
                    pass
        return dataframe


class PandasDataset(Dataset[PandasModel]):
    ...


class ListOfPandasDatasetsWithSameNumberOfFiles(Model[List[PandasDataset]]):
    @classmethod
    def _parse_data(cls, dataset_list: List[PandasDataset]) -> Any:
        assert len(dataset_list) >= 2
        assert all(len(dataset) for dataset in dataset_list)



from io import BytesIO
from typing import Any, Dict, IO, Type, Union

from omnipy.data.serializer import TarFileSerializer

from . import pd
from ...data.dataset import Dataset
from .models import PandasDataset


class PandasDatasetToTarFileSerializer(TarFileSerializer):
    @classmethod
    def get_supported_dataset_type(cls) -> Type[Dataset]:
        return PandasDataset

    @classmethod
    def get_output_file_suffix(cls) -> str:
        return 'csv'

    @classmethod
    def serialize(cls, pandas_dataset: PandasDataset) -> Union[bytes, memoryview]:
        def pandas_encode_func(pandas_data: pd.DataFrame) -> memoryview:
            csv_bytes = BytesIO()
            pandas_data.to_csv(csv_bytes, encoding='utf8', mode='b', index=False)
            return csv_bytes.getbuffer()

        return cls.create_tarfile_from_dataset(pandas_dataset, data_encode_func=pandas_encode_func)

    @classmethod
    def deserialize(cls, tarfile_bytes: bytes) -> PandasDataset:
        pandas_dataset = PandasDataset()

        def csv_decode_func(file_stream: IO[bytes]) -> pd.DataFrame:
            return pd.read_csv(file_stream, encoding='utf8')

        def python_dictify_object(obj_type: str, obj_val: Any) -> Dict:
            return {obj_type: obj_val}

        cls.create_dataset_from_tarfile(
            pandas_dataset,
            tarfile_bytes,
            data_decode_func=csv_decode_func,
            dictify_object_func=python_dictify_object,
            import_method='from_data')  # noqa

        return pandas_dataset



from abc import ABC, abstractmethod
from typing import Optional, Type

from omnipy.api.protocols import IsEngineConfig, IsRunStateRegistry


class Engine(ABC):
    def __init__(self) -> None:
        config_cls = self.get_config_cls()
        self._config: IsEngineConfig = config_cls()
        self._registry: Optional[IsRunStateRegistry] = None

        self._init_engine()

    @abstractmethod
    def _init_engine(self) -> None:
        """
        Private method for initialization of an Engine subclass. Removes the need to override the
        __init__() method in the subclass, reducing class coupling and possible error sources.
        Must be implemented by all subclasses of Engine, but can remain empty.
        """

    @abstractmethod
    def _update_from_config(self) -> None:
        """
        Private method to signal to subclass to update any internal state that depends on config
        values.
        Must be implemented by all subclasses of Engine, but can remain empty.
        """

    @classmethod
    @abstractmethod
    def get_config_cls(cls) -> Type[IsEngineConfig]:
        """
        Specification of config class mapped to an Engine subclass. Must be implemented by all
        subclasses of Engine. If no configuration is needed, the EngineConfig class can be returned.
        :return: Class implementing the IsEngineConfig protocol
        """

    def set_config(self, config: IsEngineConfig) -> None:
        self._config = config
        self._update_from_config()

    def set_registry(self, registry: Optional[IsRunStateRegistry]) -> None:
        self._registry = registry



from io import StringIO
import os

from omnipy.compute.task import TaskTemplate
from omnipy.modules.raw.protocols import (IsModifyAllLinesCallable,
                                          IsModifyContentsCallable,
                                          IsModifyEachLineCallable)


@TaskTemplate(iterate_over_data_files=True)
def modify_datafile_contents(
    data_file: str,
    modify_contents_func: IsModifyContentsCallable,
    **kwargs: object,
) -> str:
    return modify_contents_func(data_file, **kwargs)


@TaskTemplate(iterate_over_data_files=True)
def modify_each_line(
    data_file: str,
    modify_line_func: IsModifyEachLineCallable,
    **kwargs: object,
) -> str:
    output_data = StringIO()
    for i, line in enumerate(StringIO(data_file)):
        modified_line = modify_line_func(i, line, **kwargs)
        if modified_line is not None:
            output_data.write(modified_line)
    return output_data.getvalue()


@TaskTemplate(iterate_over_data_files=True)
def modify_all_lines(
    data_file: str,
    modify_all_lines_func: IsModifyAllLinesCallable,
    **kwargs: object,
) -> str:
    all_lines = [line.strip() for line in StringIO(data_file)]
    modified_lines = modify_all_lines_func(all_lines, **kwargs)
    return os.linesep.join(modified_lines)



from __future__ import annotations

from enum import Enum, IntEnum


class PersistOutputsOptions(str, Enum):
    DISABLED = 'disabled'
    FOLLOW_CONFIG = 'config'
    ENABLED = 'enabled'


class RestoreOutputsOptions(str, Enum):
    DISABLED = 'disabled'
    FOLLOW_CONFIG = 'config'
    AUTO_ENABLE_IGNORE_PARAMS = 'auto_ignore_params'
    FORCE_ENABLE_IGNORE_PARAMS = 'force_ignore_params'


class ConfigPersistOutputsOptions(str, Enum):
    DISABLED = 'disabled'
    ENABLE_FLOW_OUTPUTS = 'flow'
    ENABLE_FLOW_AND_TASK_OUTPUTS = 'all'


class ConfigRestoreOutputsOptions(str, Enum):
    DISABLED = 'disabled'
    AUTO_ENABLE_IGNORE_PARAMS = 'auto_ignore_params'


class EngineChoice(str, Enum):
    LOCAL = 'local'
    PREFECT = 'prefect'


class RunState(IntEnum):
    INITIALIZED = 1
    RUNNING = 2
    FINISHED = 3


# TODO: Add 'apply' state
# TODO: Add 'failed' state and error management
# TODO: Consider the need for a 'waiting' state


class RunStateLogMessages(str, Enum):
    INITIALIZED = 'Initialized "{}"'
    RUNNING = 'Started running "{}"...'
    FINISHED = 'Finished running "{}"!'



import json
from types import NoneType
from typing import Any, Generic, get_args, get_origin, Type, TypeVar, Union

# from orjson import orjson
from pydantic import Protocol, root_validator
from pydantic.fields import ModelField, Undefined
from pydantic.generics import GenericModel

RootT = TypeVar('RootT')
ROOT_KEY = '__root__'

# def orjson_dumps(v, *, default):
#     # orjson.dumps returns bytes, to match standard json.dumps we need to decode
#     return orjson.dumps(v, default=default).decode()


class Model(GenericModel, Generic[RootT]):
    """
    Model is a generic class. Subclasses of Model need to specify the type of
    contents that is accepted according to its model. Example:

    class MyNumberList(DatasetModel[List[int]]):
        pass

    See also docs of the Dataset class for more usage examples.
    """

    __root__: RootT

    class Config:
        arbitrary_types_allowed = True
        validate_all = True
        validate_assignment = True
        smart_union = True
        # json_loads = orjson.loads
        # json_dumps = orjson_dumps

    @classmethod
    def _get_default_value_from_model(cls, model: Type[RootT]) -> RootT:
        if isinstance(model, TypeVar):
            if model.__bound__ is None:  # noqa
                raise TypeError('The TypeVar "{}" needs to be bound to a '.format(model.__name__)
                                + 'type that provides a default value when called')
            else:
                model = model.__bound__  # noqa
        origin_type = get_origin(model)
        args = get_args(model)

        if origin_type in (None, ()):
            origin_type = model

        if origin_type is Union:
            if NoneType in args:
                return None
            for arg in args:
                if callable(arg):
                    return cls._get_default_value_from_model(arg)
        elif origin_type is tuple:
            if args and Ellipsis not in args:
                return tuple(cls._get_default_value_from_model(arg) for arg in args)

        return origin_type()

    @classmethod
    def _populate_root_field(cls, model: Type[RootT]) -> None:
        default_val = cls._get_default_value_from_model(model)

        def get_default_val() -> RootT:
            return default_val

        if ROOT_KEY in cls.__config__.fields:
            cls.__config__.fields[ROOT_KEY]['default_factory'] = get_default_val
        else:
            cls.__config__.fields[ROOT_KEY] = {'default_factory': get_default_val}

        data_field = ModelField.infer(
            name=ROOT_KEY,
            value=Undefined,
            annotation=model,
            class_validators=None,
            config=cls.__config__)

        cls.__fields__[ROOT_KEY] = data_field
        cls.__annotations__[ROOT_KEY] = model

    @classmethod
    def _depopulate_root_field(cls):
        del cls.__config__.fields[ROOT_KEY]
        del cls.__fields__[ROOT_KEY]
        del cls.__annotations__[ROOT_KEY]

    def __class_getitem__(cls, model: Union[Type[RootT], TypeVar]) -> Union[Type[RootT], TypeVar]:
        # TODO: change model type to params: Union[Type[Any], Tuple[Type[Any], ...]]
        #       as in GenericModel

        # For now, only singular model types are allowed. These lines are needed for
        # interoperability with pydantic GenericModel, which internally stores the model
        # as a tuple:
        if isinstance(model, tuple) and len(model) == 1:
            model = model[0]

        # Populating the root field at runtime instead of providing a __root__ Field explicitly
        # is needed due to the inability of typing/pydantic to provide a dynamic default based on
        # the actual type. The following issue in mypy seems relevant:
        # https://github.com/python/mypy/issues/3737 (as well as linked issues)
        if cls == Model:  # Only for the base Model class
            cls._populate_root_field(model)

        created_model = super().__class_getitem__(model)

        # As long as models are not created concurrently, setting the class members temporarily
        # should not have averse effects
        # TODO: Check if we can move to explicit definition of __root__ field at the object
        #       level in pydantic 2.0 (when it is released)
        if cls == Model:
            cls._depopulate_root_field()

        return created_model

    def __new__(cls, value=Undefined, **kwargs):
        model_not_specified = ROOT_KEY not in cls.__fields__
        if model_not_specified:
            cls._raise_no_model_exception()
        return super().__new__(cls)

    def __init__(self, value=Undefined, /, **data: RootT) -> None:
        if value is not Undefined:
            data[ROOT_KEY] = value

        super().__init__(**data)

        if not self.__class__.__doc__:
            self._set_standard_field_description()

    @staticmethod
    def _raise_no_model_exception() -> None:
        raise TypeError('Note: The Model class requires a concrete model to be specified as '
                        'a type hierarchy within brackets either directly, e.g.:\n\n'
                        '\tmodel = Model[List[int]]([1,2,3])\n\n'
                        'or indirectly in a subclass definition, e.g.:\n\n'
                        '\tclass MyNumberList(Model[List[int]]): ...\n\n')

    def _set_standard_field_description(self) -> None:
        self.__fields__[ROOT_KEY].field_info.description = self._get_standard_field_description()

    @classmethod
    def _get_standard_field_description(cls) -> str:
        return ('This class represents a concrete model for data items in the `omnipy` Python '
                'package. It is a statically typed specialization of the Model class, '
                'which is itself wrapping the excellent Python package named `pydantic`.')

    @classmethod
    def _parse_data(cls, data: RootT) -> Any:
        return data

    @root_validator
    def _parse_root_object(cls, root_obj: RootT) -> Any:  # noqa
        if ROOT_KEY not in root_obj:
            return root_obj
        else:
            return {ROOT_KEY: cls._parse_data(root_obj[ROOT_KEY])}

    @property
    def contents(self) -> Any:
        return self.__dict__.get(ROOT_KEY)

    @contents.setter
    def contents(self, value: RootT) -> None:
        super().__setattr__(ROOT_KEY, value)

    def to_data(self) -> Any:
        return self.dict()[ROOT_KEY]

    def from_data(self, value: Any) -> None:
        self.contents = value

    def to_json(self, pretty=False) -> str:
        json_content = self.json()
        if pretty:
            return self._pretty_print_json(json.loads(json_content))
        else:
            return json_content

    def from_json(self, json_contents: str) -> None:
        new_model = self.parse_raw(json_contents, proto=Protocol.json)
        self._set_contents_without_validation(new_model)

    # @classmethod
    # def get_type_args(cls):
    #     return cls.__fields__.get(ROOT_KEY).type_
    #
    # @classmethod
    # def create_from_json(cls, data: Union[str, Tuple[str]]):
    #     if isinstance(data, tuple):
    #         data = data[0]
    #
    #     obj = cls()
    #     obj.from_json(data)
    #     return obj
    #
    # def __reduce__(self):
    #     return self.__class__.create_from_json, (self.to_json(),)

    def _set_contents_without_validation(self, contents: RootT) -> None:
        validate_assignment = self.__config__.validate_assignment
        self.__config__.validate_assignment = False
        self.contents = contents.contents
        self.__config__.validate_assignment = validate_assignment

    @classmethod
    def to_json_schema(cls, pretty=False) -> str:
        schema = cls.schema()

        if pretty:
            return cls._pretty_print_json(schema)
        else:
            return json.dumps(schema)

    @staticmethod
    def _pretty_print_json(json_content: Any) -> str:
        return json.dumps(json_content, indent=4)

    def _check_for_root_key(self) -> None:
        if ROOT_KEY not in self.__dict__:
            raise TypeError('The Model class requires the specific model to be specified in as '
                            'a type hierarchy within brackets either directly, e.g.:\n'
                            '\t"model = Model[List[int]]([1,2,3])"\n'
                            'or indirectly in a subclass definition, e.g.:\n'
                            '\t"class MyNumberList(Model[List[int]]): ..."')

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in self.__dict__ and attr not in [ROOT_KEY]:
            super().__setattr__(attr, value)
        else:
            if attr in ['contents']:
                contents_prop = getattr(self.__class__, attr)
                contents_prop.__set__(self, value)
            else:
                raise RuntimeError('Model does not allow setting of extra attributes')



from abc import ABC, abstractmethod
from io import BytesIO
import tarfile
from tarfile import TarInfo
from typing import Any, Callable, Dict, IO, Tuple, Type, Union

from pydantic import ValidationError

from omnipy.data.dataset import Dataset


class Serializer(ABC):
    @classmethod
    @abstractmethod
    def get_supported_dataset_type(cls) -> Type[Dataset]:
        pass

    @classmethod
    @abstractmethod
    def get_output_file_suffix(cls) -> str:
        pass

    @classmethod
    @abstractmethod
    def serialize(cls, dataset: Dataset) -> Union[bytes, memoryview]:
        pass

    @classmethod
    @abstractmethod
    def deserialize(cls, serialized: bytes) -> Dataset:
        pass


class TarFileSerializer(Serializer, ABC):
    @classmethod
    def create_tarfile_from_dataset(cls,
                                    dataset: Dataset,
                                    data_encode_func: Callable[[Any], Union[bytes, memoryview]]):
        bytes_io = BytesIO()
        with tarfile.open(fileobj=bytes_io, mode='w:gz') as tarfile_stream:
            for obj_type, data_obj in dataset.items():
                json_data_bytestream = BytesIO(data_encode_func(data_obj))
                json_data_bytestream.seek(0)
                tarinfo = TarInfo(name=f'{obj_type}.{cls.get_output_file_suffix()}')
                tarinfo.size = len(json_data_bytestream.getbuffer())
                tarfile_stream.addfile(tarinfo, json_data_bytestream)
        return bytes_io.getbuffer().tobytes()

    @classmethod
    def create_dataset_from_tarfile(cls,
                                    dataset: Dataset,
                                    tarfile_bytes: bytes,
                                    data_decode_func: Callable[[IO[bytes]], Any],
                                    dictify_object_func: Callable[[str, Any], Union[Dict, str]],
                                    import_method='from_data'):
        with tarfile.open(fileobj=BytesIO(tarfile_bytes), mode='r:gz') as tarfile_stream:
            for filename in tarfile_stream.getnames():
                obj_type_file = tarfile_stream.extractfile(filename)
                assert filename.endswith(f'.{cls.get_output_file_suffix()}')
                obj_type = '.'.join(filename.split('.')[:-1])
                getattr(dataset, import_method)(
                    dictify_object_func(obj_type, data_decode_func(obj_type_file)))


class SerializerRegistry:
    def __init__(self) -> None:
        self._serializer_classes: list[Type[Serializer]] = []

    def register(self, serializer_cls: Type[Serializer]) -> None:
        self._serializer_classes.append(serializer_cls)

    @property
    def serializers(self) -> Tuple[Type[Serializer], ...]:
        return tuple(self._serializer_classes)

    @property
    def tar_file_serializers(self) -> Tuple[Type[TarFileSerializer], ...]:
        return tuple(cls for cls in self._serializer_classes if issubclass(cls, TarFileSerializer))

    def auto_detect(self, dataset: Dataset):
        return self._autodetect_serializer(dataset, self.serializers)

    def auto_detect_tar_file_serializer(self, dataset: Dataset):
        return self._autodetect_serializer(dataset, self.tar_file_serializers)

    @classmethod
    def _autodetect_serializer(cls, dataset, serializers):
        # def _direct(dataset, new_dataset_cls):
        #     new_dataset = new_dataset_cls(dataset)
        #     return new_dataset

        def _to_data_from_json(dataset, new_dataset_cls):
            new_dataset = new_dataset_cls()
            new_dataset.from_json(dataset.to_data())
            return new_dataset

        def _to_data_from_data(dataset, new_dataset_cls):
            new_dataset = new_dataset_cls()
            new_dataset.from_data(dataset.to_data())
            return new_dataset

        # def _to_json_from_json(dataset, new_dataset_cls):
        #     new_dataset = new_dataset_cls()
        #     new_dataset.from_json(dataset.to_json())
        #     return new_dataset

        for func in (_to_data_from_json, _to_data_from_data):
            for serializer in serializers:
                new_dataset_cls = serializer.get_supported_dataset_type()

                try:
                    new_dataset = func(dataset, new_dataset_cls)
                    return new_dataset, serializer
                except (TypeError, ValueError, ValidationError) as e:
                    pass

        return None, None

    def detect_tar_file_serializers_from_file_suffix(self, file_suffix: str):
        return tuple(serializer_cls for serializer_cls in self.tar_file_serializers
                     if serializer_cls.get_output_file_suffix() == file_suffix)



from typing import Dict, IO, Type, Union

from omnipy.data.serializer import TarFileSerializer

from ...data.dataset import Dataset
from .models import JsonDataset, JsonModel


class JsonDatasetToTarFileSerializer(TarFileSerializer):
    @classmethod
    def get_supported_dataset_type(cls) -> Type[Dataset]:
        return JsonDataset

    @classmethod
    def get_output_file_suffix(cls) -> str:
        return 'json'

    @classmethod
    def serialize(cls, json_dataset: JsonDataset) -> Union[bytes, memoryview]:
        def json_encode_func(json_data: JsonModel) -> bytes:
            return json_data.json(indent=2).encode('utf8')

        return cls.create_tarfile_from_dataset(json_dataset, data_encode_func=json_encode_func)

    @classmethod
    def deserialize(cls, tarfile_bytes: bytes) -> JsonDataset:
        json_dataset = JsonDataset()

        def json_decode_func(file_stream: IO[bytes]) -> str:
            return file_stream.read().decode('utf8')

        def json_dictify_object(obj_type: str, obj_val: str) -> Dict[str, str]:
            return {f'{obj_type}': f'{obj_val}'}

        cls.create_dataset_from_tarfile(
            json_dataset,
            tarfile_bytes,
            data_decode_func=json_decode_func,
            dictify_object_func=json_dictify_object,
            import_method='from_json')

        return json_dataset



from datetime import datetime

from omnipy.api.protocols import IsNestedContext


class FlowContextJobMixin:
    def __init__(self) -> None:
        self._time_of_last_run = None

    @property
    def flow_context(self) -> IsNestedContext:
        class FlowContext:
            @classmethod
            def __enter__(cls):
                self.__class__.job_creator.__enter__()
                self._time_of_last_run = self.time_of_cur_toplevel_flow_run

            @classmethod
            def __exit__(cls, exc_type, exc_val, exc_tb):
                self.__class__.job_creator.__exit__(exc_type, exc_val, exc_tb)

        return FlowContext()

    @property
    def time_of_last_run(self) -> datetime:
        return self._time_of_last_run



from collections import defaultdict
from datetime import datetime
import logging
from typing import DefaultDict, Dict, List, Optional, Tuple, Union

from omnipy.api.constants import OMNIPY_LOG_FORMAT_STR
from omnipy.api.enums import RunState, RunStateLogMessages
from omnipy.api.protocols import IsJob, IsRunStateRegistryConfig
from omnipy.config.registry import RunStateRegistryConfig
from omnipy.util.helpers import get_datetime_format


class RunStateRegistry:
    def __init__(self) -> None:
        self._datetime_format: Optional[str] = None
        self._logger: Optional[logging.Logger] = None
        self._config: IsRunStateRegistryConfig = RunStateRegistryConfig()

        self._jobs: Dict[str, IsJob] = {}
        self._job_states: Dict[str, RunState] = {}
        self._state_jobs: DefaultDict[RunState, List[str]] = defaultdict(list)
        self._job_state_datetime: Dict[Tuple[str, RunState], datetime] = {}

    def get_job_state(self, job: IsJob) -> RunState:
        return self._job_states[job.unique_name]

    def get_job_state_datetime(self, job: IsJob, state: RunState) -> datetime:
        return self._job_state_datetime[(job.unique_name, state)]

    def all_jobs(self, state: Optional[RunState] = None) -> Tuple[IsJob, ...]:
        if state is not None:
            job_unique_names = self._state_jobs[state]
            return tuple(self._jobs[unique_name] for unique_name in job_unique_names)
        else:
            return tuple(self._jobs.values())

    def set_logger(self,
                   logger: Optional[logging.Logger],
                   set_omnipy_formatter_on_handlers=True,
                   locale: Union[str, Tuple[str, str]] = '') -> None:

        self._logger = logger
        if self._logger is not None:
            self._datetime_format = get_datetime_format(locale)

            if set_omnipy_formatter_on_handlers:
                formatter = logging.Formatter(OMNIPY_LOG_FORMAT_STR)
                for handler in self._logger.handlers:
                    handler.setFormatter(formatter)

    def set_config(self, config: IsRunStateRegistryConfig) -> None:
        self._config = config

    def set_job_state(self, job: IsJob, state: RunState) -> None:
        cur_datetime = datetime.now()

        if job.unique_name in self._jobs:
            self._update_job_registration(job, state)
        else:
            self._register_new_job(job, state)

        self._update_job_stats(job, state, cur_datetime)
        self._log_state_change(job, state)

    def _other_job_registered_with_same_unique_name(self, job: IsJob) -> bool:
        other_job_same_unique_name = self._jobs.get(job.unique_name)
        return other_job_same_unique_name and id(other_job_same_unique_name) != id(job)

    def _update_job_registration(self, job: IsJob, state: RunState) -> None:
        # TODO: Reimplement logic using a state machine, e.g. "transitions" package
        if self._other_job_registered_with_same_unique_name(job):
            while self._other_job_registered_with_same_unique_name(job):
                job.regenerate_unique_name()
            self._register_new_job(job, state)
        else:
            prev_state = self._job_states[job.unique_name]
            if state == prev_state + 1:
                self._state_jobs[prev_state].remove(job.unique_name)
            else:
                self._raise_job_error(
                    job,
                    f'Transitioning from state {prev_state.name} '
                    f'to state {state.name} is not allowed',
                )

    def _register_new_job(self, job, state) -> None:
        if state != RunState.INITIALIZED:
            self._raise_job_error(
                job,
                f'Initial state of must be "INITIALIZED", not "{state.name}"',
            )
        self._jobs[job.unique_name] = job

    def _update_job_stats(self, job, state, cur_datetime) -> None:
        self._job_states[job.unique_name] = state
        self._state_jobs[state].append(job.unique_name)
        self._job_state_datetime[(job.unique_name, state)] = cur_datetime

    def _log_state_change(self, job: IsJob, state: RunState) -> None:
        if self._logger is not None:
            log_msg = RunStateLogMessages[state.name].value.format(job.unique_name)
            datetime_obj = self.get_job_state_datetime(job, state)
            self.log(datetime_obj, log_msg)

    def log(self, datetime_obj: datetime, log_msg: str):
        if self._logger is not None:
            datetime_str = datetime_obj.strftime(self._datetime_format)
            self._logger.info(f'{datetime_str}: {log_msg}')

    def _raise_job_error(self, job: IsJob, msg: str) -> None:
        raise ValueError(f'Error in job "{job.unique_name}": {msg}')
